#  Copyright 2021 United Kingdom Research and Innovation
#  Copyright 2021 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt

import unittest
from cil.framework import AcquisitionGeometry
from cil.utilities.dataexample import SIMULATED_PARALLEL_BEAM_DATA, SIMULATED_CONE_BEAM_DATA, SIMULATED_SPHERE_VOLUME
from scipy.fft  import fft, ifft
from skimage.transform.radon_transform import _get_fourier_filter as skimage_get_fourier_filter
import numpy as np
from utils import has_tigre, has_ipp, has_astra, has_nvidia, has_matplotlib, has_skimage, initialise_tests

from cil.recon.Reconstructor import Reconstructor # checks on baseclass
from cil.recon.FBP import GenericFilteredBackProjection # checks on baseclass
from cil.recon import FDK, FBP

import os, sys

from scipy.fft import fftfreq
import tempfile

initialise_tests()

if has_tigre:
    from cil.plugins.tigre import ProjectionOperator as ProjectionOperator_tigre
    from cil.plugins.tigre import FBP as FBP_tigre
    from tigre.utilities.filtering import ramp_flat, filter

if has_astra:
    from cil.plugins.astra import ProjectionOperator as ProjectionOperator_astra
    from cil.plugins.astra import FBP as FBP_astra

if has_matplotlib:
    import matplotlib.testing.compare as compare

if has_skimage:
    from skimage.transform.radon_transform import _get_fourier_filter as skimage_get_fourier_filter

@unittest.skipUnless(has_tigre, "TIGRE not installed")
class Test_Reconstructor_tigre(unittest.TestCase):

    def setUp(self):
        #%% Setup Geometry
        voxel_num_xy = 255
        voxel_num_z = 15

        mag = 2
        src_to_obj = 50
        src_to_det = src_to_obj * mag

        pix_size = 0.2
        det_pix_x = voxel_num_xy
        det_pix_y = voxel_num_z

        num_projections = 1000
        angles = np.linspace(0, 360, num=num_projections, endpoint=False)

        self.ag = AcquisitionGeometry.create_Cone2D([0,-src_to_obj],[0,src_to_det-src_to_obj])\
                                           .set_angles(angles)\
                                           .set_panel(det_pix_x, pix_size)\
                                           .set_labels(['angle','horizontal'])

        self.ig = self.ag.get_ImageGeometry()

        self.ag3D = AcquisitionGeometry.create_Cone3D([0,-src_to_obj,0],[0,src_to_det-src_to_obj,0])\
                                     .set_angles(angles)\
                                     .set_panel((det_pix_x,det_pix_y), (pix_size,pix_size))\
                                     .set_labels(['angle','vertical','horizontal'])
        self.ig3D = self.ag3D.get_ImageGeometry()

        self.ad3D = self.ag3D.allocate('random', seed=3)
        self.ig3D = self.ag3D.get_ImageGeometry()

        self.ad2D = self.ag.allocate('random', seed=3)

    def test_defaults(self):
        reconstructor = Reconstructor(self.ad3D)
        self.assertEqual(id(reconstructor.input),id(self.ad3D))
        self.assertEqual(reconstructor.image_geometry,self.ig3D)
        self.assertEqual(reconstructor.backend, 'tigre')

    def test_set_input(self):
        reconstructor = Reconstructor(self.ad3D)
        self.assertEqual(id(reconstructor.input),id(self.ad3D))

        ag3D_new = self.ad3D.copy()
        reconstructor.set_input(ag3D_new)
        self.assertEqual(id(reconstructor.input),id(ag3D_new))

        ag3D_new = self.ad3D.get_slice(vertical='centre')
        with self.assertRaises(ValueError):
            reconstructor.set_input(ag3D_new)

        with self.assertRaises(TypeError):
            reconstructor = Reconstructor(self.ag3D)

    def test_weak_input(self):

        data = self.ad3D.copy()
        reconstructor = Reconstructor(data)
        self.assertEqual(id(reconstructor.input),id(data))
        del data

        with self.assertRaises(ValueError):
            reconstructor.input

        reconstructor.set_input(self.ad3D)
        self.assertEqual(id(reconstructor.input),id(self.ad3D))

    def test_set_image_data(self):
        reconstructor = Reconstructor(self.ad3D)

        self.ig3D.voxel_num_z = 1
        reconstructor.set_image_geometry(self.ig3D)
        self.assertEqual(reconstructor.image_geometry,self.ig3D)

    def test_set_backend(self):

        with self.assertRaises(ValueError):
            reconstructor = Reconstructor(self.ad3D, backend='unsupported_backend')


@unittest.skipUnless(has_tigre and has_ipp, "TIGRE or IPP not installed")
class Test_GenericFilteredBackProjection_tigre(unittest.TestCase):

    def setUp(self):
        #%% Setup Geometry
        voxel_num_xy = 16
        voxel_num_z = 4

        mag = 2
        src_to_obj = 50
        src_to_det = src_to_obj * mag

        pix_size = 0.2
        det_pix_x = voxel_num_xy
        det_pix_y = voxel_num_z

        num_projections = 36
        angles = np.linspace(0, 360, num=num_projections, endpoint=False)

        self.ag = AcquisitionGeometry.create_Cone2D([0,-src_to_obj],[0,src_to_det-src_to_obj])\
                                           .set_angles(angles)\
                                           .set_panel(det_pix_x, pix_size)\
                                           .set_labels(['angle','horizontal'])

        self.ig = self.ag.get_ImageGeometry()

        self.ag3D = AcquisitionGeometry.create_Cone3D([0,-src_to_obj,0],[0,src_to_det-src_to_obj,0])\
                                     .set_angles(angles)\
                                     .set_panel((det_pix_x,det_pix_y), (pix_size,pix_size))\
                                     .set_labels(['angle','vertical','horizontal'])
        self.ig3D = self.ag3D.get_ImageGeometry()

        self.ad3D = self.ag3D.allocate('random', seed=4)
        self.ig3D = self.ag3D.get_ImageGeometry()

    def check_defaults(self, reconstructor):
        self.assertEqual(reconstructor.filter, 'ram-lak')
        self.assertEqual(reconstructor.fft_order, 8)
        self.assertFalse(reconstructor.filter_inplace)
        self.assertIsNone(reconstructor._weights)

        filter = reconstructor.get_filter_array()
        self.assertEqual(type(filter), np.ndarray)
        self.assertEqual(len(filter), 2**8)
        self.assertEqual(filter[0], 0)
        self.assertEqual(filter[128],1.0)
        self.assertEqual(filter[1],filter[255])

        self.assertEqual(reconstructor.image_geometry,self.ig3D)

    def test_defaults(self):

        reconstructor = GenericFilteredBackProjection(self.ad3D)
        self.check_defaults(reconstructor)

    def test_reset(self):
        reconstructor = GenericFilteredBackProjection(self.ad3D)
        reconstructor.set_fft_order(10)

        arr = reconstructor.get_filter_array()
        arr.fill(0)
        reconstructor.set_filter(arr)

        ig = self.ig3D.copy()
        ig.num_voxels_x = 4
        reconstructor.set_image_geometry(ig)

        reconstructor.set_filter_inplace(True)

        reconstructor.reset()
        self.check_defaults(reconstructor)

    def test_set_filter(self):
        reconstructor = GenericFilteredBackProjection(self.ad3D)

        with self.assertRaises(ValueError):
            reconstructor.set_filter("unsupported_filter")

        # test all supported filters are set
        for x in reconstructor.preset_filters:
            reconstructor.set_filter(x)
            self.assertEqual(reconstructor.filter, x, msg='Mismatch on test: Filter {0}'.format(x))
            self.assertEqual(reconstructor._filter_cutoff, 1.0, msg='Mismatch on test: Filter {0}'.format(x))

        # test filter cut-off is set
        reconstructor.set_filter('ram-lak', 0.5)
        self.assertEqual(reconstructor._filter_cutoff, 0.5,msg='Filter cut-off frequency mismatch')

        # test custom array is set
        filter = reconstructor.get_filter_array()
        filter_new =filter *0.5
        reconstructor.set_filter(filter_new)
        self.assertEqual(reconstructor.filter, 'custom')
        filter = reconstructor.get_filter_array()
        np.testing.assert_array_equal(filter,filter_new)

        with self.assertRaises(ValueError):
            reconstructor.set_filter(filter[1:-1])

    @unittest.skipUnless(has_skimage, "SKIMAGE not installed")
    def test_get_filter_array(self):

        reconstructor = GenericFilteredBackProjection(self.ad3D)

        #filters constructed in different domains but at higher orders this bias is negligible
        order = 20
        reconstructor.set_fft_order(order)

        reconstructor.set_filter(filter='ram-lak', cutoff=1.0)
        arr = reconstructor.get_filter_array()
        response = skimage_get_fourier_filter(2**order, 'ramp')
        np.testing.assert_almost_equal(arr, response[:,0], 6, "Failed with filter 'ram-lak'")

        reconstructor.set_filter(filter='ram-lak', cutoff=1.0)
        arr = reconstructor.get_filter_array()
        response = skimage_get_fourier_filter(2**order, 'ramp')
        response[response>1.0]=0
        np.testing.assert_almost_equal(arr, response[:,0], 6, "Failed with filter 'ram-lak' and cut-off frequency")

        filters = ['shepp-logan', 'cosine', 'hamming', 'hann']

        for filter in filters:
            reconstructor.set_fft_order(order)
            reconstructor.set_filter(filter=filter, cutoff=1.0)
            arr = reconstructor.get_filter_array()

            response = skimage_get_fourier_filter(2**order, filter)
            np.testing.assert_almost_equal(arr, response[:,0], 6, "Failed with filter {}".format(filter))

    def test_set_fft_order(self):
        reconstructor = GenericFilteredBackProjection(self.ad3D)
        reconstructor.set_fft_order(10)
        self.assertEqual(reconstructor.fft_order, 10)

        with self.assertRaises(ValueError):
            reconstructor.set_fft_order(2)

    def test_set_filter_inplace(self):
        reconstructor = GenericFilteredBackProjection(self.ad3D)
        reconstructor.set_filter_inplace(True)
        self.assertTrue(reconstructor.filter_inplace)

        with self.assertRaises(TypeError):
            reconstructor.set_filter_inplace('unsupported_value')

    def create_custom_filter_example(self, cutoff):
        """Returns a custom filter array."""
        filter_length = 256
        freq = fftfreq(filter_length)
        freq *= 2
        ramp = abs(freq)
        ramp[ramp>cutoff] = 0
        FBP_filter = ramp*(np.cos(freq*np.pi*4)+1*np.cos(1/5*freq*np.pi/2))/2
        return FBP_filter

    @unittest.skipUnless(has_matplotlib, "matplotlib not installed")
    def test_plot_filter(self):
        """
        Tests that the filters are plotted correctly for two different
        values of cutoff. This is done for all preset filters and the custom filter.
        The plots are compared to stored png files.
        The test will not show any screen output.
        The temporary directory and files are removed.
        """
        fdk = GenericFilteredBackProjection(self.ad3D)
        filter_list = fdk.preset_filters
        filter_list.append('custom')
        filter_plots_folder = os.path.join(os.path.dirname(__file__),"test_plots","filters")
        test_plot_folder = tempfile.mkdtemp(suffix=None, prefix=None, dir=None)
        test_plot_path = os.path.join(test_plot_folder, 'test_plot_filter.png')
        for cutoff in [0.5,1]:
            for filter_name in filter_list:
                if filter_name == 'custom':
                    FBP_filter = self.create_custom_filter_example(cutoff)
                else:
                    FBP_filter =filter_name
                fdk.set_filter(FBP_filter, cutoff)
                plot = fdk.plot_filter()
                base_plot_path =os.path.join(filter_plots_folder, filter_name+'_'+str(round(cutoff))+'.png')
                plot.savefig(test_plot_path)
                err = compare.compare_images(base_plot_path,  test_plot_path, tol=0)
                self.assertIsNone(err, f"Filter plots are not the same: {err}")
                os.remove(test_plot_path)
                plot.close()
        os.removedirs(test_plot_folder)


@unittest.skipUnless(has_tigre and has_ipp, "TIGRE or IPP not installed")
class Test_FDK_tigre(unittest.TestCase):

    def setUp(self):
        #%% Setup Geometry
        voxel_num_xy = 16
        voxel_num_z = 4

        mag = 2
        src_to_obj = 50
        src_to_det = src_to_obj * mag

        pix_size = 0.2
        det_pix_x = voxel_num_xy
        det_pix_y = voxel_num_z

        num_projections = 36
        angles = np.linspace(0, 360, num=num_projections, endpoint=False)

        self.ag = AcquisitionGeometry.create_Cone2D([0,-src_to_obj],[0,src_to_det-src_to_obj])\
                                           .set_angles(angles)\
                                           .set_panel(det_pix_x, pix_size)\
                                           .set_labels(['angle','horizontal'])

        self.ig = self.ag.get_ImageGeometry()

        self.ag3D = AcquisitionGeometry.create_Cone3D([0,-src_to_obj,0],[0,src_to_det-src_to_obj,0])\
                                     .set_angles(angles)\
                                     .set_panel((det_pix_x,det_pix_y), (pix_size,pix_size))\
                                     .set_labels(['angle','vertical','horizontal'])
        self.ig3D = self.ag3D.get_ImageGeometry()

        self.ad3D = self.ag3D.allocate('random', seed=5)
        self.ig3D = self.ag3D.get_ImageGeometry()

    def test_set_filter(self):

        reconstructor = FDK(self.ad3D)
        filter = reconstructor.get_filter_array()
        filter_new =filter *0.5
        reconstructor.set_filter(filter_new)

        reconstructor.set_fft_order(10)
        with self.assertRaises(ValueError):
            reconstructor._pre_filtering(self.ad3D)

        with self.assertRaises(ValueError):
            reconstructor.set_filter(filter[1:-1])

    def test_filtering(self):
        ag = AcquisitionGeometry.create_Cone3D([0,-1,0],[0,2,0])\
            .set_panel([64,3],[0.1,0.1])\
            .set_angles([0,90])

        ad = ag.allocate('random',seed=0)

        reconstructor = FDK(ad)
        out1 = ad.copy()
        reconstructor._pre_filtering(out1)

        #by hand
        filter = reconstructor.get_filter_array()
        reconstructor._calculate_weights(ag)
        pad0 = (len(filter)-ag.pixel_num_h)//2
        pad1 = len(filter)-ag.pixel_num_h-pad0

        out2 = ad.array.copy()
        out2*=reconstructor._weights
        for i in range(2):
            proj_padded = np.zeros((ag.pixel_num_v,len(filter)))
            proj_padded[:,pad0:-pad1] = out2[i]
            filtered_proj=fft(proj_padded,axis=-1)
            filtered_proj*=filter
            filtered_proj=ifft(filtered_proj,axis=-1)
            out2[i]=np.real(filtered_proj)[:,pad0:-pad1]

        diff = (out1-out2).abs().max()
        self.assertLess(diff, 1e-5)

    def test_weights(self):
        ag = AcquisitionGeometry.create_Cone3D([0,-1,0],[0,2,0])\
            .set_panel([3,4],[0.1,0.2])\
            .set_angles([0,90])
        ad = ag.allocate(0)

        reconstructor = FDK(ad)
        reconstructor._calculate_weights(ag)
        weights = reconstructor._weights

        scaling =  7.5 * np.pi
        weights_new = np.ones_like(weights)

        det_size_x = ag.pixel_size_h*ag.pixel_num_h
        det_size_y = ag.pixel_size_v*ag.pixel_num_v

        ray_length_z = 3
        for j in range(4):
            ray_length_y = -det_size_y/2 +  ag.pixel_size_v * (j+0.5)
            for i in range(3):
                ray_length_x = -det_size_x/2 +  ag.pixel_size_h * (i+0.5)
                ray_length = (ray_length_x**2+ray_length_y**2+ray_length_z**2)**0.5
                weights_new[j,i] = scaling*ray_length_z/ray_length

        diff = np.max(np.abs(weights - weights_new))
        self.assertLess(diff, 1e-5)


@unittest.skipUnless(has_tigre and has_ipp, "TIGRE or IPP not installed")
class Test_FBP_tigre(unittest.TestCase):

    def setUp(self):
        #%% Setup Geometry
        voxel_num_xy = 16
        voxel_num_z = 4

        pix_size = 0.2
        det_pix_x = voxel_num_xy
        det_pix_y = voxel_num_z

        num_projections = 36
        angles = np.linspace(0, 360, num=num_projections, endpoint=False)

        self.ag = AcquisitionGeometry.create_Parallel2D()\
                                           .set_angles(angles)\
                                           .set_panel(det_pix_x, pix_size)\
                                           .set_labels(['angle','horizontal'])

        self.ig = self.ag.get_ImageGeometry()

        self.ag3D = AcquisitionGeometry.create_Parallel3D()\
                                     .set_angles(angles)\
                                     .set_panel((det_pix_x,det_pix_y), (pix_size,pix_size))\
                                     .set_labels(['angle','vertical','horizontal'])
        self.ig3D = self.ag3D.get_ImageGeometry()

        self.ad3D = self.ag3D.allocate('random', seed=3)
        self.ig3D = self.ag3D.get_ImageGeometry()

    def test_set_filter(self):
        reconstructor = FBP(self.ad3D)
        filter = reconstructor.get_filter_array()
        filter_new =filter *0.5
        reconstructor.set_filter(filter_new)

        reconstructor.set_fft_order(10)
        with self.assertRaises(ValueError):
            reconstructor._pre_filtering(self.ad3D)

        with self.assertRaises(ValueError):
            reconstructor.set_filter(filter[1:-1])

    def test_split_processing(self):
        reconstructor = FBP(self.ad3D)

        self.assertEqual(reconstructor.slices_per_chunk, 0)

        reconstructor.set_split_processing(1)
        self.assertEqual(reconstructor.slices_per_chunk, 1)

        reconstructor.reset()
        self.assertEqual(reconstructor.slices_per_chunk, 0)

    def test_filtering(self):
        ag = AcquisitionGeometry.create_Parallel3D()\
            .set_panel([64,3],[0.1,0.1])\
            .set_angles([0,90])

        ad = ag.allocate('random',seed=0)

        reconstructor = FBP(ad)
        out1 = ad.copy()
        reconstructor._pre_filtering(out1)

        #by hand
        filter = reconstructor.get_filter_array()
        reconstructor._calculate_weights(ag)
        pad0 = (len(filter)-ag.pixel_num_h)//2
        pad1 = len(filter)-ag.pixel_num_h-pad0

        out2 = ad.array.copy()
        out2*=reconstructor._weights
        for i in range(2):
            proj_padded = np.zeros((ag.pixel_num_v,len(filter)))
            proj_padded[:,pad0:-pad1] = out2[i]
            filtered_proj=fft(proj_padded,axis=-1)
            filtered_proj*=filter
            filtered_proj=ifft(filtered_proj,axis=-1)
            out2[i]=np.real(filtered_proj)[:,pad0:-pad1]

        diff = (out1-out2).abs().max()
        self.assertLess(diff, 1e-5)

    def test_weights(self):
        ag = AcquisitionGeometry.create_Parallel3D()\
            .set_panel([3,4],[0.1,0.2])\
            .set_angles([0,90])
        ad = ag.allocate(0)

        reconstructor = FBP(ad)
        reconstructor._calculate_weights(ag)
        weights = reconstructor._weights

        scaling =  (2 * np.pi/ ag.num_projections) / ( 4 * ag.pixel_size_h )
        weights_new = np.ones_like(weights) * scaling

        np.testing.assert_allclose(weights,weights_new)

    def test_set_backend(self):

        ag = AcquisitionGeometry.create_Parallel3D()\
            .set_panel([3,4],[0.1,0.2])\
            .set_angles([0,90])\
            .set_labels(['vertical','angle','horizontal'])
        ad = ag.allocate(0)

        reconstructor = FBP(ad, backend='astra')

        #Reconstructor backend raises an error if backend isn't supported
        with self.assertRaises(ValueError):
            reconstructor = FBP(ad, backend='unsupported_backend')

        #Reconstructor backend raises an error if dataorder isn't compatible
        with self.assertRaises(ValueError):
            reconstructor = FBP(ad, backend='tigre')


@unittest.skipUnless(has_tigre and has_nvidia and has_ipp, "TIGRE or IPP not installed")
class Test_FDK_results_tigre_ipp(unittest.TestCase):
    def setUp(self):

        self.acq_data = SIMULATED_CONE_BEAM_DATA.get()
        self.img_data = SIMULATED_SPHERE_VOLUME.get()

        self.acq_data=np.log(self.acq_data)
        self.acq_data*=-1.0

        self.ig = self.img_data.geometry
        self.ag = self.acq_data.geometry

    def test_results_3D(self):

        reconstructor = FDK(self.acq_data)

        reco = reconstructor.run(verbose=0)
        np.testing.assert_allclose(reco.as_array(), self.img_data.as_array(),atol=1e-3)

        reco2 = reco.copy()
        reco2.fill(0)
        reconstructor.run(out=reco2, verbose=0)
        np.testing.assert_allclose(reco.as_array(), reco2.as_array(), atol=1e-8)

    def test_results_2D(self):

        data2D = self.acq_data.get_slice(vertical='centre')
        img_data2D = self.img_data.get_slice(vertical='centre')

        reconstructor = FDK(data2D)
        reco = reconstructor.run(verbose=0)
        np.testing.assert_allclose(reco.as_array(), img_data2D.as_array(),atol=1e-3)

        reco2 = reco.copy()
        reco2.fill(0)
        reconstructor.run(out=reco2, verbose=0)
        np.testing.assert_allclose(reco.as_array(), reco2.as_array(), atol=1e-8)

    def test_results_with_tigre(self):

        fbp_tigre = FBP_tigre(self.ig, self.ag)
        reco_tigre = fbp_tigre(self.acq_data)

        #fbp CIL with TIGRE's filter
        reconstructor_cil = FDK(self.acq_data)
        n = 2**reconstructor_cil.fft_order
        ramp = ramp_flat(n)
        filt = filter('ram_lak',ramp[0],n,1,False)

        reconstructor_cil = FDK(self.acq_data)
        reconstructor_cil.set_filter(filt)
        reco_cil = reconstructor_cil.run(verbose=0)

        #with the same filter results should be virtually identical
        np.testing.assert_allclose(reco_cil.as_array(), reco_tigre.as_array(),atol=1e-8)

    def test_results_inplace_filtering(self):

        reconstructor = FDK(self.acq_data)
        reco = reconstructor.run(verbose=0)

        data_filtered= self.acq_data.copy()
        reconstructor_inplace = FDK(data_filtered)
        reconstructor_inplace.set_filter_inplace(True)
        reconstructor_inplace.run(out=reco, verbose=0)

        diff = (data_filtered - self.acq_data).abs().mean()
        self.assertGreater(diff,0.8)


@unittest.skipUnless(has_tigre and has_nvidia and has_ipp, "TIGRE or IPP not installed")
class Test_FBP_tigre_ipp(unittest.TestCase):
    def setUp(self):
        self.acq_data = SIMULATED_PARALLEL_BEAM_DATA.get()
        self.img_data = SIMULATED_SPHERE_VOLUME.get()

        self.acq_data=np.log(self.acq_data)
        self.acq_data*=-1.0

        self.ig = self.img_data.geometry
        self.ag = self.acq_data.geometry

    def test_results_3D(self):
        reconstructor = FBP(self.acq_data)

        reco = reconstructor.run(verbose=0)
        np.testing.assert_allclose(reco.as_array(), self.img_data.as_array(),atol=1e-3)

        reco2 = reco.copy()
        reco2.fill(0)
        reconstructor.run(out=reco2, verbose=0)
        np.testing.assert_allclose(reco.as_array(), reco2.as_array(), atol=1e-8)

    def test_results_3D_split(self):

        reconstructor = FBP(self.acq_data)
        reconstructor.set_split_processing(8)

        reco = reconstructor.run(verbose=0)
        np.testing.assert_allclose(reco.as_array(), self.img_data.as_array(),atol=1e-3)

        reco2 = reco.copy()
        reco2.fill(0)
        reconstructor.run(out=reco2, verbose=0)
        np.testing.assert_allclose(reco.as_array(), reco2.as_array(), atol=1e-8)

    def test_results_3D_split_reverse(self):
        acq_data = self.acq_data.copy()
        acq_data.geometry.config.panel.origin = 'top-left'

        reconstructor = FBP(acq_data)
        reconstructor.set_split_processing(8)

        expected_image = np.flip(self.img_data.as_array(),0)

        reco = reconstructor.run(verbose=0)
        np.testing.assert_allclose(reco.as_array(), expected_image,atol=1e-3)

        reco2 = reco.copy()
        reco2.fill(0)
        reconstructor.run(out=reco2, verbose=0)
        np.testing.assert_allclose(reco.as_array(), reco2.as_array(), atol=1e-8)

    def test_results_2D(self):
        data2D = self.acq_data.get_slice(vertical='centre')
        img_data2D = self.img_data.get_slice(vertical='centre')

        reconstructor = FBP(data2D)
        reco = reconstructor.run(verbose=0)
        np.testing.assert_allclose(reco.as_array(), img_data2D.as_array(),atol=1e-3)

        reco2 = reco.copy()
        reco2.fill(0)
        reconstructor.run(out=reco2, verbose=0)
        np.testing.assert_allclose(reco.as_array(), reco2.as_array(), atol=1e-8)

    def test_results_with_tigre(self):
        fbp_tigre = FBP_tigre(self.ig, self.ag)
        reco_tigre = fbp_tigre(self.acq_data)

        #fbp CIL with TIGRE's filter
        reconstructor_cil = FBP(self.acq_data)
        n = 2**reconstructor_cil.fft_order
        ramp = ramp_flat(n)
        filt = filter('ram_lak',ramp[0],n,1,False)

        reconstructor_cil = FBP(self.acq_data)
        reconstructor_cil.set_filter(filt)
        reco_cil = reconstructor_cil.run(verbose=0)

        #with the same filter results should be virtually identical
        np.testing.assert_allclose(reco_cil.as_array(), reco_tigre.as_array(),atol=1e-8)

    def test_results_inplace_filtering(self):
        reconstructor = FBP(self.acq_data)
        reco = reconstructor.run(verbose=0)

        data_filtered= self.acq_data.copy()
        reconstructor_inplace = FBP(data_filtered)
        reconstructor_inplace.set_filter_inplace(True)
        reconstructor_inplace.run(out=reco, verbose=0)

        diff = (data_filtered - self.acq_data).abs().mean()
        self.assertGreater(diff,0.8)


@unittest.skipUnless(has_astra and has_nvidia and has_ipp, "ASTRA or IPP not installed")
class Test_FBP_astra_ipp(unittest.TestCase):
    def setUp(self):
        self.acq_data = SIMULATED_PARALLEL_BEAM_DATA.get()
        self.img_data = SIMULATED_SPHERE_VOLUME.get()

        self.acq_data=np.log(self.acq_data)
        self.acq_data*=-1.0
        self.acq_data.reorder('astra')

        self.ig = self.img_data.geometry
        self.ag = self.acq_data.geometry

    def test_results_3D(self):
        reconstructor = FBP(self.acq_data, backend='astra')

        reco = reconstructor.run(verbose=0)
        np.testing.assert_allclose(reco.as_array(), self.img_data.as_array(),atol=1e-3)

        reco2 = reco.copy()
        reco2.fill(0)
        reconstructor.run(out=reco2, verbose=0)
        np.testing.assert_allclose(reco.as_array(), reco2.as_array(), atol=1e-8)

    def test_results_3D_split(self):

        reconstructor = FBP(self.acq_data, backend='astra')
        reconstructor.set_split_processing(8)

        reco = reconstructor.run(verbose=0)
        np.testing.assert_allclose(reco.as_array(), self.img_data.as_array(),atol=1e-3)

        reco2 = reco.copy()
        reco2.fill(0)
        reconstructor.run(out=reco2, verbose=0)
        np.testing.assert_allclose(reco.as_array(), reco2.as_array(), atol=1e-8)

    def test_results_3D_split_reverse(self):
        acq_data = self.acq_data.copy()
        acq_data.geometry.config.panel.origin = 'top-left'

        reconstructor = FBP(acq_data, backend='astra')
        reconstructor.set_split_processing(8)

        expected_image = np.flip(self.img_data.as_array(),0)

        reco = reconstructor.run(verbose=0)
        np.testing.assert_allclose(reco.as_array(), expected_image,atol=1e-3)

        reco2 = reco.copy()
        reco2.fill(0)
        reconstructor.run(out=reco2, verbose=0)
        np.testing.assert_allclose(reco.as_array(), reco2.as_array(), atol=1e-8)

    def test_results_2D(self):
        data2D = self.acq_data.get_slice(vertical='centre')
        data2D.reorder('astra')
        img_data2D = self.img_data.get_slice(vertical='centre')

        reconstructor = FBP(data2D, backend='astra')
        reco = reconstructor.run(verbose=0)
        np.testing.assert_allclose(reco.as_array(), img_data2D.as_array(),atol=1e-3)

        reco2 = reco.copy()
        reco2.fill(0)
        reconstructor.run(out=reco2, verbose=0)
        np.testing.assert_allclose(reco.as_array(), reco2.as_array(), atol=1e-8)

    def test_results_with_astra(self):
        fbp_astra = FBP_astra(self.ig, self.ag)
        reco_astra = fbp_astra(self.acq_data)

        reconstructor_cil = FBP(self.acq_data, backend='astra')
        reco_cil = reconstructor_cil.run(verbose=0)

        np.testing.assert_allclose(reco_cil.as_array(), reco_astra.as_array(),atol=1e-4)

    def test_results_inplace_filtering(self):
        reconstructor = FBP(self.acq_data, backend='astra')
        reco = reconstructor.run(verbose=0)

        data_filtered= self.acq_data.copy()
        reconstructor_inplace = FBP(data_filtered, backend='astra')
        reconstructor_inplace.set_filter_inplace(True)
        reconstructor_inplace.run(out=reco, verbose=0)

        diff = (data_filtered - self.acq_data).abs().mean()
        self.assertGreater(diff,0.8)



#  Copyright 2018 United Kingdom Research and Innovation
#  Copyright 2018 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt

from .Algorithm import Algorithm
from .CGLS import CGLS
from .SIRT import SIRT
from .GD import GD
from .FISTA import FISTA
from .FISTA import ISTA
from .FISTA import ISTA as PGD
from .APGD import APGD 
from .PDHG import PDHG
from .ADMM import LADMM
from .SPDHG import SPDHG
from .PD3O import PD3O



#  Copyright 2019 United Kingdom Research and Innovation
#  Copyright 2019 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt

import warnings
from utils import has_cvxpy
import unittest
from os import unlink
from tempfile import NamedTemporaryFile

import numpy as np
import logging


from cil.framework import VectorData, ImageData, ImageGeometry, AcquisitionData, AcquisitionGeometry, BlockDataContainer, BlockGeometry

from cil.framework.labels import FillType

from cil.optimisation.utilities import ArmijoStepSizeRule, ConstantStepSize, Sampler, callbacks, Sensitivity, StepSizeRule
from cil.optimisation.algorithms.APGD import NesterovMomentum, ScalarMomentumCoefficient, ConstantMomentum
from cil.optimisation.operators import IdentityOperator, AdjointOperator
from cil.optimisation.operators import GradientOperator, BlockOperator, MatrixOperator


from cil.optimisation.functions import MixedL21Norm, BlockFunction, L1Norm, KullbackLeibler, IndicatorBox, LeastSquares, ZeroFunction, L2NormSquared, OperatorCompositionFunction, TotalVariation, SGFunction, SVRGFunction, SAGAFunction, SAGFunction, LSVRGFunction, ScaledFunction
from cil.optimisation.algorithms import Algorithm, GD, CGLS, SIRT, FISTA, ISTA, SPDHG, PDHG, LADMM, PD3O, PGD, APGD


from scipy.optimize import minimize, rosen

from cil.utilities import dataexample
from cil.utilities import noise as applynoise
from cil.optimisation.functions import Rosenbrock
from cil.utilities.quality_measures import mae, mse, psnr

import logging
from testclass import CCPiTestClass
from utils import has_astra

from unittest.mock import MagicMock

log = logging.getLogger(__name__)


if has_cvxpy:
    import cvxpy

from unittest.mock import MagicMock, patch
class TestGD(CCPiTestClass):
    def setUp(self):

        x0_1 = 1.1
        x0_2 = 1.1
        # x0_1 = 0.5
        # x0_2 = 0.5
        self.x0 = np.array([x0_1, x0_2])

        self.initial = VectorData(np.array(self.x0))
        method = 'Nelder-Mead'  # or "BFGS"
        # self.scipy_opt_low = minimize(rosen, self.x0, method=method, tol=1e-3, options={"maxiter":50})
        self.scipy_opt_high = minimize(
            rosen, self.x0, method=method, tol=1e-2)  # (1., 1.)
        # fixed (alpha=1, beta=100) same to Scipy, min at (alpha,alpha^2)
        self.f = Rosenbrock(alpha=1, beta=100)

    def test_GD(self):
        ig = ImageGeometry(12, 13, 14)
        initial = ig.allocate()
        b = ig.allocate('random', seed=3)
        identity = IdentityOperator(ig)

        norm2sq = LeastSquares(identity, b)

        step_size = norm2sq.L / 3.

        alg = GD(initial=initial, f=norm2sq, step_size=step_size,
                 atol=1e-9, rtol=1e-6)
        alg.run(1000, verbose=0)
        self.assertNumpyArrayAlmostEqual(alg.x.as_array(), b.as_array())

        alg = GD(initial=initial, f=norm2sq, step_size=step_size,
                 atol=1e-9, rtol=1e-6, update_objective_interval=2)
        self.assertTrue(alg.update_objective_interval == 2)
        alg.run(20, verbose=0)
        self.assertNumpyArrayAlmostEqual(alg.x.as_array(), b.as_array())

        self.assertAlmostEqual(alg.get_last_objective(), 0)
        self.assertNotAlmostEqual(alg.loss[0], alg.loss[1])

    def test_update_interval_0(self):
        '''
        Checks that an algorithm runs with no problems when
        the update_objective interval is set to 0 and with
        verbose on / off
        '''
        ig = ImageGeometry(12, 13, 14)
        initial = ig.allocate()
        b = ig.allocate('random', seed=3)
        identity = IdentityOperator(ig)
        norm2sq = LeastSquares(identity, b)
        alg = GD(initial=initial,
                 f=norm2sq,
                 update_objective_interval=0,
                 atol=1e-9, rtol=1e-6)
        self.assertTrue(alg.update_objective_interval == 0)
        alg.run(20, verbose=True)
        self.assertNumpyArrayAlmostEqual(alg.x.as_array(), b.as_array())
        alg.run(20, verbose=False)
        self.assertNumpyArrayAlmostEqual(alg.x.as_array(), b.as_array())

    def test_gd_step_size_init(self):
        gd = GD(initial=self.initial, f=self.f, step_size=0.002)
        self.assertEqual(gd.step_size_rule.step_size, 0.002)
        self.assertEqual(gd.step_size, 0.002)

        gd = GD(initial=self.initial, f=self.f)
        self.assertEqual(gd.step_size_rule.alpha_orig, 1e6)
        self.assertEqual(gd.step_size_rule.beta, 0.5)
        self.assertEqual(gd.step_size_rule.max_iterations, np.ceil(
            2 * np.log10(1e6) / np.log10(2)))
        with self.assertRaises(NotImplementedError):
            self.assertEqual(gd.step_size,3)

        gd = GD(initial=self.initial,
                f=self.f, alpha=1e2, beta=0.25)
        self.assertEqual(gd.step_size_rule.alpha_orig, 1e2)
        self.assertEqual(gd.step_size_rule.beta, 0.25)
        self.assertEqual(gd.step_size_rule.max_iterations, np.ceil(
            2 * np.log10(1e2) / np.log10(2)))


    def test_gd_constant_step_size_init(self):
        rule = ConstantStepSize(0.4)
        self.assertEqual(rule.step_size, 0.4)
        gd = GD(initial=self.initial,
                f=self.f, step_size=rule)
        self.assertEqual(gd.step_size_rule.step_size, 0.4)
        self.assertEqual(gd.step_size, 0.4)

    def test_gd_fixed_step_size_rosen(self):

        gd = GD(initial=self.initial, f=self.f, step_size=0.002,
                update_objective_interval=500)
        gd.run(3000, verbose=0)
        np.testing.assert_allclose(
            gd.solution.array[0], self.scipy_opt_high.x[0], atol=1e-2)
        np.testing.assert_allclose(
            gd.solution.array[1], self.scipy_opt_high.x[1], atol=1e-2)

    def test_armijo_step_size_init(self):

        rule = ArmijoStepSizeRule()
        self.assertEqual(rule.alpha_orig, 1e6)
        self.assertEqual(rule.beta, 0.5)
        self.assertEqual(rule.max_iterations, np.ceil(
            2 * np.log10(1e6) / np.log10(2)))

        gd = GD(initial=self.initial,
                f=self.f, step_size=rule)
        self.assertEqual(gd.step_size_rule.alpha_orig, 1e6)
        self.assertEqual(gd.step_size_rule.beta, 0.5)
        self.assertEqual(gd.step_size_rule.max_iterations, np.ceil(
            2 * np.log10(1e6) / np.log10(2)))

        rule = ArmijoStepSizeRule(5e5, 0.2, 5)
        self.assertEqual(rule.alpha_orig, 5e5)
        self.assertEqual(rule.beta, 0.2)
        self.assertEqual(rule.max_iterations, 5)

        gd = GD(initial=self.initial,
                f=self.f, step_size=rule)
        self.assertEqual(gd.step_size_rule.alpha_orig, 5e5)
        self.assertEqual(gd.step_size_rule.beta, 0.2)
        self.assertEqual(gd.step_size_rule.max_iterations, 5)

        with self.assertRaises(NotImplementedError):
            gd.step_size

    def test_GDArmijo(self):
        ig = ImageGeometry(12, 13, 14)
        initial = ig.allocate()
        # b = initial.copy()
        # fill with random numbers
        # b.fill(np.random.random(initial.shape))
        b = ig.allocate('random', seed=3)
        identity = IdentityOperator(ig)

        norm2sq = LeastSquares(identity, b)

        alg = GD(initial=initial, f=norm2sq)
        alg.run(100, verbose=0)
        self.assertNumpyArrayAlmostEqual(alg.x.as_array(), b.as_array())
        alg = GD(initial=initial, f=norm2sq, update_objective_interval=2)
        self.assertTrue(alg.update_objective_interval==2)

        alg = GD(initial=initial, f=norm2sq,
                 update_objective_interval=2)
        self.assertTrue(alg.update_objective_interval == 2)

        alg.run(20, verbose=0)
        self.assertNumpyArrayAlmostEqual(alg.x.as_array(), b.as_array())

    def test_gd_armijo_rosen(self):
        armj = ArmijoStepSizeRule(alpha=50, max_iterations=50, warmstart=False)
        gd = GD(initial=self.initial, f=self.f, step_size=armj,
                update_objective_interval=500)
        gd.run(3500, verbose=0)
        np.testing.assert_allclose(
            gd.solution.array[0], self.scipy_opt_high.x[0], atol=1e-2)
        np.testing.assert_allclose(
            gd.solution.array[1], self.scipy_opt_high.x[1], atol=1e-2)

    def test_gd_run_no_iterations(self):
        gd = GD(initial=self.initial, f=self.f, step_size=0.002)
        with self.assertRaises(ValueError):
            gd.run()

    def test_gd_run_infinite(self):
        gd = GD(initial=self.initial, f=self.f, step_size=0.002)
        with self.assertRaises(ValueError):
            gd.run(np.inf)

        class StopCallback(callbacks.Callback):
            def __init__(self):
                self.count = 0

            def __call__(self, algorithm):
                self.count += 1
                if self.count == 10:
                    raise StopIteration
        with self.assertWarns(UserWarning):
            gd.run(np.inf, callbacks=[StopCallback()])
        self.assertEqual(gd.iteration, 9)

    def test_gd_deprecate_atol_rtol(self):
        with self.assertWarns(DeprecationWarning):
            initial=VectorData(np.array([1.1,1.1]))
            b=VectorData(np.array([1.,1.]))
            A=IdentityOperator(b.geometry)
            gd = GD(initial=initial, f=LeastSquares(A,b ), step_size=1, atol=1, rtol=2)
            
        self.assertEqual(gd.rtol, 2.)
        self.assertEqual(gd.atol, 1.)
        gd.run(10)
        self.assertEqual(gd.iteration, 0)
        
class Test_APGD(CCPiTestClass):
    def setUp(self):
        self.ig = ImageGeometry(11,12,13)
        self.initial = self.ig.allocate(0)
        self.b = self.ig.allocate(None)
        self.b.fill(np.array(range(11*12*13)).reshape(13,12,11))
        self.identity = IdentityOperator(self.ig)

        self.f = OperatorCompositionFunction(L2NormSquared(b=self.b), self.identity)
        self.g= IndicatorBox(lower=0)
        
    def test_init_default(self):
        alg = APGD(initial=self.initial, f=self.f, g=self.g)
        assert isinstance(alg.momentum, NesterovMomentum) 
        assert isinstance(alg.step_size_rule, ConstantStepSize)
        self.assertEqual(alg.step_size_rule.step_size, 1/self.f.L)
        self.assertEqual(alg.update_objective_interval, 1)
        self.assertNumpyArrayAlmostEqual(alg.y.as_array(), self.initial.as_array())

    def test_init_momentum(self):
        alg = APGD(initial=self.initial, f=self.f, g=self.g, momentum=1)
        assert isinstance(alg.momentum, ConstantMomentum)
        self.assertEqual(alg.momentum(alg),1) 
        self.assertEqual(alg.momentum.momentum, 1)
        
        with self.assertRaises(TypeError):
            alg = APGD(initial=self.initial, f=self.f, g=self.g, momentum='banana')
        
    def test_init_step_size(self):
        alg = APGD(initial=self.initial, f=self.f, g=self.g, step_size=1.0)
        assert isinstance(alg.step_size_rule, ConstantStepSize)
        self.assertEqual(alg.step_size_rule.get_step_size(alg), 1.0)
        self.assertEqual(alg.step_size_rule.step_size, 1.0)
        
        with self.assertRaises(TypeError):
            alg = APGD(initial=self.initial, f=self.f, g=self.g, step_size='banana')

        alg = APGD(initial=self.initial, f=self.f, g=self.g)
        self.assertEqual(alg.step_size, 1/self.f.L)
        self.assertEqual(alg.step_size_rule.step_size, 1/self.f.L)
        
        class CustomStep(StepSizeRule):
            def get_step_size(self, algorithm):
                return 0.45/(algorithm.iteration+1)
            
        alg = APGD(initial=self.initial, f=self.f, g=self.g, step_size=CustomStep())
        with self.assertRaises(NotImplementedError):
           alg.step_size
        alg.run(1)
        self.assertEqual(alg.step_size, 0.45)
        self.assertEqual(alg.step_size, 0.45)
        alg.run(1)
        self.assertEqual(alg.step_size, 0.45/2)
        self.assertEqual(alg.step_size, 0.45/2)
        
        
    def test_update_step(self):
        alg = APGD(initial=self.initial, f=self.f, g=self.g, step_size=0.3, momentum=0.5)
        y_0=self.initial.copy()
        x_0=self.g.proximal(y_0 - 0.3*self.f.gradient(y_0), 0.3)
        y_1=x_0 + 0.5*(x_0-y_0)
        alg.run(1)
        self.assertNumpyArrayAlmostEqual(alg.y.as_array(), y_1.as_array())
        
        x_1=self.g.proximal(y_1 - 0.3*self.f.gradient(y_1), 0.3)
        y_2=x_1 + 0.5*(x_1-x_0)
        alg.run(1)
        self.assertNumpyArrayAlmostEqual(alg.y.as_array(), y_2.as_array())
        
    def test_provable_convergence(self):
        alg = APGD(initial=self.initial, f=self.f, g=self.g)
        self.assertTrue(alg.is_provably_convergent())
        
        alg = APGD(initial=self.initial, f=self.f, g=self.g, momentum=0.5)
        with self.assertRaises(TypeError):
            alg.is_provably_convergent()
            
        alg = APGD(initial=self.initial, f=self.f, g=self.g, preconditioner=Sensitivity(self.identity))
        with self.assertRaises(NotImplementedError):
            alg.is_provably_convergent()
        
        
        
        

class TestFISTA(CCPiTestClass):
    def test_FISTA(self):
        ig = ImageGeometry(127, 139, 149)
        initial = ig.allocate(0)
        b = initial.copy()
        # fill with random numbers
        b.fill(np.random.random(initial.shape))
        initial = ig.allocate(FillType["RANDOM"], seed=3)
        identity = IdentityOperator(ig)

        norm2sq = OperatorCompositionFunction(L2NormSquared(b=b), identity)
        opt = {'tol': 1e-4, 'memopt': False}
        log.info("initial objective %s", norm2sq(initial))

        alg = FISTA(initial=initial, f=norm2sq, g=ZeroFunction())
        alg.run(20, verbose=0)
        self.assertNumpyArrayAlmostEqual(alg.x.as_array(), b.as_array())

        alg = FISTA(initial=initial, f=norm2sq, g=ZeroFunction(),
                    update_objective_interval=2)

        self.assertTrue(alg.update_objective_interval == 2)

        alg.run(20, verbose=0)
        self.assertNumpyArrayAlmostEqual(alg.x.as_array(), b.as_array())

        # Testing g=None
        alg = FISTA(initial=initial, f=norm2sq, g=None,
                    update_objective_interval=2)
        self.assertTrue(alg.update_objective_interval == 2)
        alg.run(20, verbose=0)
        self.assertNumpyArrayAlmostEqual(alg.x.as_array(), b.as_array())

        # Testing f=None
        alg = FISTA(initial=initial, f=None, g=L1Norm(b=b),
                    update_objective_interval=2)
        self.assertTrue(alg.update_objective_interval == 2)
        alg.run(20, verbose=0)
        self.assertNumpyArrayAlmostEqual(alg.x.as_array(), b.as_array())

        # Testing f and g is None
        with self.assertRaises(ValueError):
            alg = FISTA(initial=initial, f=None, g=None,
                        update_objective_interval=2)

    def test_FISTA_update(self):

        # setup linear system to solve
        np.random.seed(10)
        n = 50
        m = 500

        A = np.random.uniform(0, 1, (m, n)).astype('float32')
        b = (A.dot(np.random.randn(n)) + 0.1 *
             np.random.randn(m)).astype('float32')

        Aop = MatrixOperator(A)
        bop = VectorData(b)

        f = LeastSquares(Aop, b=bop, c=0.5)
        g = ZeroFunction()

        ig = Aop.domain

        initial = ig.allocate()

        # ista run 10 iteration
        tmp_initial = ig.allocate()
        fista = FISTA(initial=tmp_initial, f=f, g=g)
        fista.run(1)

        # fista update method
        t_old = 1

        step_size = 1.0/f.L
        x_old = ig.allocate()
        y_old = ig.allocate()

        for _ in range(1):

            x = g.proximal(y_old - step_size *
                           f.gradient(y_old), tau=step_size)
            t = 0.5*(1 + np.sqrt(1 + 4*(t_old**2)))
            y = x + ((t_old-1)/t) * (x - x_old)

            x_old.fill(x)
            y_old.fill(y)
            t_old = t

        np.testing.assert_allclose(fista.solution.array, x.array, atol=1e-2)

        # check objective
        res1 = fista.objective[-1]
        res2 = f(x) + g(x)
        self.assertTrue(res1 == res2)

        tmp_initial = ig.allocate()
        fista1 = FISTA(initial=tmp_initial, f=f, g=g)
        self.assertTrue(fista1.is_provably_convergent())

        fista1 = FISTA(initial=tmp_initial, f=f, g=g,
                       step_size=30.0)
        self.assertFalse(fista1.is_provably_convergent())
        
            

    def test_FISTA_Norm2Sq(self):
        ig = ImageGeometry(127, 139, 149)
        b = ig.allocate(FillType["RANDOM"], seed=3)
        # fill with random numbers
        initial = ig.allocate(FillType["RANDOM"], seed=4)
        identity = IdentityOperator(ig)

        norm2sq = LeastSquares(identity, b)

        opt = {'tol': 1e-4, 'memopt': False}
        log.info("initial objective %s", norm2sq(initial))
        alg = FISTA(initial=initial, f=norm2sq, g=ZeroFunction())
        alg.run(20, verbose=0)
        self.assertNumpyArrayAlmostEqual(alg.x.as_array(), b.as_array())

        alg = FISTA(initial=initial, f=norm2sq, g=ZeroFunction(),
                    update_objective_interval=3)
        self.assertTrue(alg.update_objective_interval == 3)

        alg.run(20, verbose=0)
        self.assertNumpyArrayAlmostEqual(alg.x.as_array(), b.as_array())

    def test_FISTA_catch_Lipschitz(self):
        ig = ImageGeometry(127, 139, 149)
        initial = ImageData(geometry=ig)
        initial = ig.allocate()
        b = initial.copy()
        # fill with random numbers
        b.fill(np.random.random(initial.shape))
        initial = ig.allocate(FillType["RANDOM"], seed=3)
        identity = IdentityOperator(ig)

        norm2sq = LeastSquares(identity, b)
        log.info('Lipschitz %s', norm2sq.L)
        # norm2sq.L = None
        # norm2sq.L = 2 * norm2sq.c * identity.norm()**2
        # norm2sq = OperatorCompositionFunction(L2NormSquared(b=b), identity)
        opt = {'tol': 1e-4, 'memopt': False}
        log.info("initial objective %s", norm2sq(initial))
        with self.assertRaises(TypeError):
            alg = FISTA(initial=initial, f=L1Norm(), g=ZeroFunction())


    def test_FISTA_step_size_init(self):
        np.random.seed(10)
        n = 5
        m = 3

        A = np.random.uniform(0, 1, (m, n)).astype('float32')
        b = (A.dot(np.random.randn(n)) + 0.1 *
             np.random.randn(m)).astype('float32')

        Aop = MatrixOperator(A)
        bop = VectorData(b)

        f = LeastSquares(Aop, b=bop, c=0.5)
        g = ZeroFunction()
        ig = Aop.domain

        initial = ig.allocate()
        
        alg= FISTA(initial=initial, f=f, g=g, step_size=0.002)
        self.assertEqual(alg.step_size_rule.step_size, 0.002)
        self.assertEqual(alg.step_size, 0.002)

        alg = FISTA(initial=initial, f=f, g=g)
        self.assertEqual(alg.step_size, 1/f.L)
        self.assertEqual(alg.step_size_rule.step_size, 1/f.L)
        
        class CustomStep(StepSizeRule):
            def get_step_size(self, algorithm):
                return 0.99/(algorithm.iteration+1)
            
        alg = FISTA(initial=initial, f=f, g=g, step_size=CustomStep())
        with self.assertRaises(NotImplementedError):
           alg.step_size
        alg.run(1)
        self.assertEqual(alg.step_size, 0.99)
        self.assertEqual(alg.step_size, 0.99)
        alg.run(1)
        self.assertEqual(alg.step_size, 0.99/2)
        self.assertEqual(alg.step_size, 0.99/2)

class testISTA(CCPiTestClass):

    def setUp(self):

        np.random.seed(10)
        n = 50
        m = 500

        A = np.random.uniform(0, 1, (m, n)).astype('float32')
        b = (A.dot(np.random.randn(n)) + 0.1 *
             np.random.randn(m)).astype('float32')

        self.Aop = MatrixOperator(A)
        self.bop = VectorData(b)

        self.f = LeastSquares(self.Aop, b=self.bop, c=0.5)
        self.g = ZeroFunction()
        self.h = L1Norm()

        self.ig = self.Aop.domain

        self.initial = self.ig.allocate()

    def tearDown(self):
        pass

    def test_signature(self):

        # check required arguments (initial, f, g)
        with np.testing.assert_raises(TypeError):
            ista = ISTA(f=self.f, g=self.g)

        with np.testing.assert_raises(TypeError):
            ista = ISTA(initial=self.initial, f=self.f)

        with np.testing.assert_raises(TypeError):
            ista = ISTA(initial=self.initial, g=self.g)

        # ista no step-size
        ista = ISTA(initial=self.initial, f=self.f, g=self.g)
        np.testing.assert_equal(ista.step_size, 0.99*2./self.f.L)

        # ista step-size
        tmp_step_size = 10.
        ista = ISTA(initial=self.initial, f=self.f,
                    g=self.g, step_size=tmp_step_size)
        np.testing.assert_equal(ista.step_size, tmp_step_size)

        # check initialisation
        self.assertTrue(id(ista.x) != id(ista.initial))
        self.assertTrue(id(ista.x_old) != id(ista.initial))

    def test_update(self):

        tmp_initial = self.ig.allocate()
        ista = ISTA(initial=tmp_initial, f=self.f, g=self.g)
        ista.run(1)

        x = tmp_initial.copy()
        x_old = tmp_initial.copy()

        for _ in range(1):
            x = ista.g.proximal(x_old - (0.99*2/ista.f.L)
                                * ista.f.gradient(x_old), (1./ista.f.L))
            x_old.fill(x)

        np.testing.assert_allclose(ista.solution.array, x.array, atol=1e-2)

        # check objective
        res1 = ista.objective[-1]
        res2 = self.f(x) + self.g(x)
        self.assertTrue(res1 == res2)
        
    def test_update_pgd(self):
        
        tmp_initial = self.ig.allocate()
        pgd = PGD(initial=tmp_initial, f=self.f, g=self.g)
        pgd.run(1)

        x = tmp_initial.copy()
        x_old = tmp_initial.copy()

        for _ in range(1):
            x = pgd.g.proximal(x_old - (0.99*2/pgd.f.L)
                                * pgd.f.gradient(x_old), (1./pgd.f.L))
            x_old.fill(x)

        np.testing.assert_allclose(pgd.solution.array, x.array, atol=1e-2)

        # check objective
        res1 = pgd.objective[-1]
        res2 = self.f(x) + self.g(x)
        self.assertTrue(res1 == res2)

    def test_update_g_none(self):

        
        tmp_initial = self.ig.allocate()
        ista = ISTA(initial=tmp_initial, f=self.f, g=None)
        ista.run(1)

        x = tmp_initial.copy()
        x_old = tmp_initial.copy()

        x = ista.g.proximal(x_old - (0.99*2/ista.f.L) *
                            ista.f.gradient(x_old), (1./ista.f.L))
        
        #Check if g is None, the proximal operator is the identity and thus GD = ISTA when g is None
        x2 = x_old - (0.99*2/ista.f.L) * ista.f.gradient(x_old)

        np.testing.assert_allclose(ista.solution.array, x.array, atol=1e-2)
        np.testing.assert_allclose(x.array, x2.array, atol=1e-2)

        # check objective
        res1 = ista.objective[-1]
        res2 = self.f(x) + self.g(x)
        self.assertTrue(res1 == res2)

    def test_update_f_none(self):

        # ista run 1 iteration
        tmp_initial = self.ig.allocate()
        ista = ISTA(initial=tmp_initial, f=None, g=self.h)
        ista.run(1)

        x = tmp_initial.copy()
        x_old = tmp_initial.copy()

        for _ in range(1):
            x = ista.g.proximal(x_old, ista.step_size)
            x_old.fill(x)

        np.testing.assert_allclose(ista.solution.array, x.array, atol=1e-2)

        # check objective
        res1 = ista.objective[-1]
        res2 = self.h(x)
        self.assertTrue(res1 == res2)

    def test_f_and_g_none(self):
        tmp_initial = self.ig.allocate()
        with self.assertRaises(ValueError):
            ista = ISTA(initial=tmp_initial, f=None, g=None)

    def test_provable_condition(self):

        tmp_initial = self.ig.allocate()
        ista1 = ISTA(initial=tmp_initial, f=self.f, g=self.g)
        self.assertTrue(ista1.is_provably_convergent())

        ista1 = ISTA(initial=tmp_initial, f=self.f, g=self.g,
                     step_size=30.0)
        self.assertFalse(ista1.is_provably_convergent())

    @unittest.skipUnless(has_cvxpy, "CVXpy not installed")
    def test_with_cvxpy(self):

        ista = ISTA(initial=self.initial, f=self.f,
                    g=self.g)
        ista.run(2000, verbose=0)

        u_cvxpy = cvxpy.Variable(self.ig.shape[0])
        objective = cvxpy.Minimize(
            0.5 * cvxpy.sum_squares(self.Aop.A @ u_cvxpy - self.bop.array))
        p = cvxpy.Problem(objective)
        p.solve(verbose=True, solver=cvxpy.SCS, eps=1e-4)

        np.testing.assert_allclose(p.value, ista.objective[-1], atol=1e-3)
        np.testing.assert_allclose(
            u_cvxpy.value, ista.solution.array, atol=1e-3)

 
    def test_ISTA_PGD_alias(self):

        with patch('cil.optimisation.algorithms.ISTA.set_up', MagicMock(return_value=None)) as mock_method:

            alg = PGD(initial=self.initial, f=self.f, g=self.g, step_size=4)
            mock_method.assert_called_once_with(initial=self.initial, f=self.f, g=self.g, step_size=4, preconditioner=None)


    def test_ISTA_step_size_init(self):
        alg= ISTA(initial=self.initial, f=self.f, g=self.g, step_size=0.002)
        self.assertEqual(alg.step_size_rule.step_size, 0.002)
        self.assertEqual(alg.step_size, 0.002)

        alg = ISTA(initial=self.initial, f=self.f, g=self.g)
        self.assertEqual(alg.step_size, 0.99*2.0/self.f.L)
        self.assertEqual(alg.step_size_rule.step_size, 0.99*2.0/self.f.L)
        
        class CustomStep(StepSizeRule):
            def get_step_size(self, algorithm):
                return 0.99**algorithm.iteration
            
        alg = ISTA(initial=self.initial, f=self.f, g=self.g, step_size=CustomStep())
        with self.assertRaises(NotImplementedError):
           alg.step_size
        alg.run(1)
        self.assertEqual(alg.step_size, 0.99**0)
        self.assertEqual(alg.step_size, 0.99**0)
        alg.run(1)
        self.assertEqual(alg.step_size, 0.99**1)
        self.assertEqual(alg.step_size, 0.99**1)

class TestCGLS(CCPiTestClass):
    def setUp(self):
        self.ig = ImageGeometry(10, 2)
        np.random.seed(2)
        self.initial = self.ig.allocate(1.)
        self.data = self.ig.allocate('random', seed=3)
        self.operator = IdentityOperator(self.ig)
        self.alg = CGLS(initial=self.initial, operator=self.operator, data=self.data,
                        update_objective_interval=2)

    # can be deprecated when tolerance is deprecated in CGLS
    def test_initialization_with_default_tolerance(self):

        self.assertEqual(self.alg.tolerance, 0)

    # can be deprecated when tolerance is deprecated in CGLS
    def test_initialization_with_custom_tolerance(self):
        with self.assertWarns(DeprecationWarning):
            alg = CGLS(initial=self.initial, operator=self.operator,
                       data=self.data, tolerance=0.01)
        self.assertEqual(alg.tolerance, 0.01)

    def test_set_up(self):
        # Test the set_up method

        self.alg.set_up(initial=self.initial,
                        operator=self.operator, data=self.data)

        # Check if internal variables are set up correctly
        self.assertNumpyArrayEqual(
            self.alg.x.as_array(), self.initial.as_array())
        self.assertEqual(self.alg.operator, self.operator)
        self.assertNumpyArrayEqual(
            self.alg.r.as_array(), (self.data-self.initial).as_array())
        self.assertNumpyArrayEqual(
            self.alg.s.as_array(), (self.data-self.initial).as_array())
        self.assertNumpyArrayEqual(
            self.alg.p.as_array(), (self.data-self.initial).as_array())
        self.assertTrue(self.alg.configured)

    def test_update(self):

        self.alg.set_up(initial=self.mock_initial,
                        operator=self.mock_operator, data=self.mock_data)

        self.alg.update()

        self.mock_operator.direct.assert_called_with(
            self.mock_data, out=self.alg.q)
        self.mock_operator.adjoint.assert_called_with(
            self.alg.r, out=self.alg.s)

    def test_convergence(self):

        self.alg.run(20, verbose=0)
        self.assertNumpyArrayAlmostEqual(
            self.alg.x.as_array(), self.data.as_array())

    # can be deprecated when tolerance is deprecated in CGLS
    def test_should_stop_flag_false(self):
        # Mocking norms to ensure tolerance isn't reached
        self.alg.run(2)
        self.alg.norms = 10
        self.alg.norms0 = 99
        self.alg.tolerance = 0.1
        self.alg.normx = 0.1

        self.assertFalse(self.alg.flag())

    # can be deprecated when tolerance is deprecated in CGLS
    def test_should_stop_flag_true(self):
        # Mocking norms to ensure tolerance is reached
        self.alg.run(4)
        self.alg.norms = 1
        self.alg.norms0 = 10
        self.alg.tolerance = 0.1
        self.alg.normx = 10

        self.assertTrue(self.alg.flag())

    # can be deprecated when tolerance is deprecated in CGLS
    def test_tolerance_reached_immediately(self):
        alg = CGLS(initial=self.operator.domain_geometry().allocate(
            0), operator=self.operator, data=self.operator.domain_geometry().allocate(0))
        alg.run(2)

    def test_update_objective(self):
        # Mocking squared_norm to return a finite value

        self.alg.r = MagicMock()
        self.alg.r.squared_norm.return_value = 1.0

        self.alg.update_objective()

        # Ensure the loss list is updated
        self.assertEqual(self.alg.loss, [1.0])

    def test_update_objective_with_nan_raises_stop_iteration(self):
        # Mocking squared_norm to return NaN
        self.alg.r = MagicMock()
        self.alg.r.squared_norm.return_value = np.nan

        with self.assertRaises(StopIteration):
            self.alg.update_objective()

    def test_update(self):
        self.alg.gamma = 4

        self.alg.update()
        norm = (self.data-self.initial).squared_norm()
        alpha = 4/norm
        self.assertNumpyArrayEqual(self.alg.x.as_array(
        ), (self.initial+alpha*(self.data-self.initial)).as_array())
        self.assertNumpyArrayEqual(self.alg.r.as_array(
        ), (self.data - self.initial-alpha*(self.data-self.initial)).as_array())
        beta = ((self.data - self.initial-alpha *
                (self.data-self.initial)).norm()**2)/4
        self.assertNumpyArrayEqual(self.alg.p.as_array(), ((
            self.data - self.initial-alpha*(self.data-self.initial))+beta*(self.data-self.initial)).as_array())


class TestPDHG(CCPiTestClass):

    def setUp(self):
        self.ig = ImageGeometry(10, 10)
        self.initial = self.ig.allocate(0)
        self.data = self.ig.allocate('random')
        self.operator = IdentityOperator(self.ig)

        self.f = L2NormSquared(b=self.data)
        self.g = ZeroFunction()

        self.subsets = 10

        data = dataexample.SIMULATED_PARALLEL_BEAM_DATA.get(size=(16, 16))

        self.partitioned_data = data.partition(self.subsets, 'sequential')
        self.A_block = BlockOperator(
            *[IdentityOperator(self.partitioned_data[i].geometry) for i in range(self.subsets)])

        # block function
        self.F_block = BlockFunction(*[L2NormSquared(b=self.partitioned_data[i])
                                for i in range(self.subsets)])
        alpha = 0.025
        self.G = alpha * IndicatorBox(lower=0)
        
        self.A_block_adjoint = AdjointOperator(self.A_block)
        self.f_adjoint = L2NormSquared(b=self.data)
        
    def test_init_primal_dual(self):
        init1 = self.ig.allocate(1)
        init2 = self.ig.allocate(2)
        init_default = self.ig.allocate(0)
        pdhg = PDHG(initial=None, f=self.f, g=self.g,
                         operator=self.operator)
        self.assertNumpyArrayAlmostEqual(pdhg.x_old.as_array(), init_default.as_array())
        self.assertNumpyArrayAlmostEqual(pdhg.y.as_array(), init_default.as_array())
        
        pdhg = PDHG(initial=init1, f=self.f, g=self.g,
                         operator=self.operator)
        
        self.assertNumpyArrayAlmostEqual(pdhg.x_old.as_array(), init1.as_array())
        self.assertNumpyArrayAlmostEqual(pdhg.y.as_array(), init_default.as_array())
        
        pdhg = PDHG(initial=(None, init2), f=self.f, g=self.g,
                         operator=self.operator) # check tuple initialisation just dual
        
        self.assertNumpyArrayAlmostEqual(pdhg.x_old.as_array(), init_default.as_array())
        self.assertNumpyArrayAlmostEqual(pdhg.y.as_array(), init2.as_array())
        
        pdhg= PDHG(initial=self.partitioned_data, f=self.f_adjoint, g=self.G,
                         operator=self.A_block_adjoint) # check can initialise with block data without dual initialisation
        self.assertNumpyArrayAlmostEqual(pdhg.x_old.get_item(0).as_array(), self.partitioned_data.get_item(0).as_array())
        self.assertEqual(np.sum(pdhg.y.as_array()), 0)
        
        pdhg = PDHG(initial=[init1, None], f=self.f, g=self.g,
                         operator=self.operator)  # check list initialisation just primal
        
        self.assertNumpyArrayAlmostEqual(pdhg.x_old.as_array(), init1.as_array())
        self.assertNumpyArrayAlmostEqual(pdhg.y.as_array(), init_default.as_array())
        
        pdhg = PDHG(initial=[init1, init2], f=self.f, g=self.g,
                         operator=self.operator)  # check list initialisation both primal and dual
        
        self.assertNumpyArrayAlmostEqual(pdhg.x_old.as_array(), init1.as_array())
        self.assertNumpyArrayAlmostEqual(pdhg.y.as_array(), init2.as_array())
        
        
        
    def test_PDHG_Denoising(self):
        # adapted from demo PDHG_TV_Color_Denoising.py in CIL-Demos repository
        data = dataexample.PEPPERS.get(size=(256, 256))
        ig = data.geometry
        ag = ig

        which_noise = 0
        # Create noisy data.
        noises = ['gaussian', 'poisson', 's&p']
        dnoise = noises[which_noise]

        def setup(data, dnoise):
            if dnoise == 's&p':
                n1 = applynoise.saltnpepper(
                    data, salt_vs_pepper=0.9, amount=0.2, seed=10)
            elif dnoise == 'poisson':
                scale = 5
                n1 = applynoise.poisson(data.as_array()/scale, seed=10)*scale
            elif dnoise == 'gaussian':
                n1 = applynoise.gaussian(data.as_array(), seed=10)
            else:
                raise ValueError('Unsupported Noise ', noise)
            noisy_data = ig.allocate()
            noisy_data.fill(n1)

            # Regularisation Parameter depending on the noise distribution
            if dnoise == 's&p':
                alpha = 0.8
            elif dnoise == 'poisson':
                alpha = 1
            elif dnoise == 'gaussian':
                alpha = .3
                # fidelity
            if dnoise == 's&p':
                g = L1Norm(b=noisy_data)
            elif dnoise == 'poisson':
                g = KullbackLeibler(b=noisy_data)
            elif dnoise == 'gaussian':
                g = 0.5 * L2NormSquared(b=noisy_data)
            return noisy_data, alpha, g

        noisy_data, alpha, g = setup(data, dnoise)
        operator = GradientOperator(
            ig, correlation=GradientOperator.CORRELATION_SPACE, backend='numpy')

        f1 = alpha * MixedL21Norm()

        # Compute operator Norm
        normK = operator.norm()

        # Primal & dual stepsizes
        sigma = 1
        tau = 1/(sigma*normK**2)

        # Setup and run the PDHG algorithm
        pdhg1 = PDHG(f=f1, g=g, operator=operator, tau=tau, sigma=sigma)
        pdhg1.update_objective_interval = 200
        pdhg1.run(1000, verbose=0)

        rmse = (pdhg1.get_output() - data).norm() / data.as_array().size
        log.info("RMSE %F", rmse)
        self.assertLess(rmse, 2e-4)

        which_noise = 1
        noise = noises[which_noise]
        noisy_data, alpha, g = setup(data, noise)
        operator = GradientOperator(
            ig, correlation=GradientOperator.CORRELATION_SPACE, backend='numpy')

        f1 = alpha * MixedL21Norm()

        # Compute operator Norm
        normK = operator.norm()

        # Primal & dual stepsizes
        sigma = 1
        tau = 1/(sigma*normK**2)

        # Setup and run the PDHG algorithm
        pdhg1 = PDHG(f=f1, g=g, operator=operator, tau=tau, sigma=sigma,
                     update_objective_interval=200)

        pdhg1.run(1000, verbose=0)

        rmse = (pdhg1.get_output() - data).norm() / data.as_array().size
        log.info("RMSE %f", rmse)
        self.assertLess(rmse, 2e-4)

        which_noise = 2
        noise = noises[which_noise]
        noisy_data, alpha, g = setup(data, noise)
        operator = GradientOperator(
            ig, correlation=GradientOperator.CORRELATION_SPACE, backend='numpy')

        f1 = alpha * MixedL21Norm()

        # Compute operator Norm
        normK = operator.norm()

        # Primal & dual stepsizes
        sigma = 1
        tau = 1/(sigma*normK**2)

        # Setup and run the PDHG algorithm
        pdhg1 = PDHG(f=f1, g=g, operator=operator, tau=tau, sigma=sigma)
        pdhg1.update_objective_interval = 200
        pdhg1.run(1000, verbose=0)

        rmse = (pdhg1.get_output() - data).norm() / data.as_array().size
        log.info("RMSE %f", rmse)
        self.assertLess(rmse, 2e-4)

    def test_PDHG_step_sizes(self):
        ig = ImageGeometry(3, 3)
        data = ig.allocate('random', seed=3)

        f = L2NormSquared(b=data)
        g = L2NormSquared()
        operator = 3*IdentityOperator(ig)

        # check if sigma, tau are None
        pdhg = PDHG(f=f, g=g, operator=operator)
        self.assertAlmostEqual(pdhg.sigma, 1./operator.norm())
        self.assertAlmostEqual(pdhg.tau, 1./operator.norm())

        # check if sigma is negative
        with self.assertRaises(ValueError):
            pdhg = PDHG(f=f, g=g, operator=operator,
                        sigma=-1)

        # check if tau is negative
        with self.assertRaises(ValueError):
            pdhg = PDHG(f=f, g=g, operator=operator, tau=-1)

        # check if tau is None
        sigma = 3.0
        pdhg = PDHG(f=f, g=g, operator=operator, sigma=sigma)
        self.assertAlmostEqual(pdhg.sigma, sigma)
        self.assertAlmostEqual(pdhg.tau, 1./(sigma * operator.norm()**2))

        # check if sigma is None
        tau = 3.0
        pdhg = PDHG(f=f, g=g, operator=operator, tau=tau)
        self.assertAlmostEqual(pdhg.tau, tau)
        self.assertAlmostEqual(pdhg.sigma, 1./(tau * operator.norm()**2))

        # check if sigma/tau are not None
        tau = 1.0
        sigma = 1.0
        pdhg = PDHG(f=f, g=g, operator=operator, tau=tau,
                    sigma=sigma)
        self.assertAlmostEqual(pdhg.tau, tau)
        self.assertAlmostEqual(pdhg.sigma, sigma)

        # check sigma/tau as arrays, sigma wrong shape
        ig1 = ImageGeometry(2, 2)
        sigma = ig1.allocate()
        with self.assertRaises(ValueError):
            pdhg = PDHG(f=f, g=g, operator=operator,
                        sigma=sigma)

        # check sigma/tau as arrays, tau wrong shape
        tau = ig1.allocate()
        with self.assertRaises(ValueError):
            pdhg = PDHG(f=f, g=g, operator=operator, tau=tau)

        # check sigma not Number or object with correct shape
        with self.assertRaises(AttributeError):
            pdhg = PDHG(f=f, g=g, operator=operator,
                        sigma="sigma")

        # check tau not Number or object with correct shape
        with self.assertRaises(AttributeError):
            pdhg = PDHG(f=f, g=g, operator=operator,
                        tau="tau")

        # check warning message if condition is not satisfied
        sigma = 4/operator.norm()
        tau = 1/3
        with self.assertWarnsRegex(UserWarning, "Convergence criterion"):
            pdhg = PDHG(f=f, g=g, operator=operator, tau=tau,
                        sigma=sigma)

        # check no warning message if check convergence is false
        sigma = 4/operator.norm()
        tau = 1/3
        with warnings.catch_warnings(record=True) as warnings_log:
            pdhg = PDHG(f=f, g=g, operator=operator, tau=tau,
                        sigma=sigma, check_convergence=False)
        self.assertEqual(warnings_log, [])

        # check no warning message if condition is satisfied
        sigma = 1/operator.norm()
        tau = 1/3
        with warnings.catch_warnings(record=True) as warnings_log:
            pdhg = PDHG(f=f, g=g, operator=operator, tau=tau,
                        sigma=sigma)
        self.assertEqual(warnings_log, [])

    def test_PDHG_strongly_convex_gamma_g(self):
        ig = ImageGeometry(3, 3)
        data = ig.allocate('random', seed=3)

        f = L2NormSquared(b=data)
        g = L2NormSquared()
        operator = IdentityOperator(ig)

        # sigma, tau
        sigma = 1.0
        tau = 1.0

        pdhg = PDHG(f=f, g=g, operator=operator, sigma=sigma, tau=tau,
                    gamma_g=0.5)
        pdhg.run(1, verbose=0)
        self.assertAlmostEqual(
            pdhg.theta, 1.0 / np.sqrt(1 + 2 * pdhg.gamma_g * tau))
        self.assertAlmostEqual(pdhg.tau, tau * pdhg.theta)
        self.assertAlmostEqual(pdhg.sigma, sigma / pdhg.theta)
        pdhg.run(4, verbose=0)
        self.assertNotEqual(pdhg.sigma, sigma)
        self.assertNotEqual(pdhg.tau, tau)

        # check negative strongly convex constant
        with self.assertRaises(ValueError):
            pdhg = PDHG(f=f, g=g, operator=operator, sigma=sigma, tau=tau,
                        gamma_g=-0.5)

        # check strongly convex constant not a number
        with self.assertRaises(ValueError):
            pdhg = PDHG(f=f, g=g, operator=operator, sigma=sigma, tau=tau,
                        gamma_g="-0.5")

    def test_PDHG_strongly_convex_gamma_fcong(self):
        ig = ImageGeometry(3, 3)
        data = ig.allocate('random', seed=3)

        f = L2NormSquared(b=data)
        g = L2NormSquared()
        operator = IdentityOperator(ig)

        # sigma, tau
        sigma = 1.0
        tau = 1.0

        pdhg = PDHG(f=f, g=g, operator=operator, sigma=sigma, tau=tau,
                    gamma_fconj=0.5)
        pdhg.run(1, verbose=0)
        self.assertEqual(pdhg.theta, 1.0 / np.sqrt(1 +
                         2 * pdhg.gamma_fconj * sigma))
        self.assertEqual(pdhg.tau, tau / pdhg.theta)
        self.assertEqual(pdhg.sigma, sigma * pdhg.theta)
        pdhg.run(4, verbose=0)
        self.assertNotEqual(pdhg.sigma, sigma)
        self.assertNotEqual(pdhg.tau, tau)

        # check negative strongly convex constant
        try:
            pdhg = PDHG(f=f, g=g, operator=operator, sigma=sigma, tau=tau,
                        gamma_fconj=-0.5)
        except ValueError as ve:
            log.info(str(ve))

        # check strongly convex constant not a number
        try:
            pdhg = PDHG(f=f, g=g, operator=operator, sigma=sigma, tau=tau,
                        gamma_fconj="-0.5")
        except ValueError as ve:
            log.info(str(ve))

    def test_PDHG_strongly_convex_both_fconj_and_g(self):

        ig = ImageGeometry(3, 3)
        data = ig.allocate('random', seed=3)

        f = L2NormSquared(b=data)
        g = L2NormSquared()
        operator = IdentityOperator(ig)

        try:
            pdhg = PDHG(f=f, g=g, operator=operator,
                        gamma_g=0.5, gamma_fconj=0.5)
            pdhg.run(verbose=0)
        except ValueError as err:
            log.info(str(err))

    def test_pdhg_theta(self):
        ig = ImageGeometry(3, 3)
        data = ig.allocate('random', seed=3)

        f = L2NormSquared(b=data)
        g = L2NormSquared()
        operator = IdentityOperator(ig)

        pdhg = PDHG(f=f, g=g, operator=operator)
        self.assertEqual(pdhg.theta, 1.0)

        pdhg = PDHG(f=f, g=g, operator=operator, theta=0.5)
        self.assertEqual(pdhg.theta, 0.5)

        with self.assertRaises(ValueError):
            PDHG(f=f, g=g, operator=operator, theta=-0.5)

        with self.assertRaises(ValueError):
            PDHG(f=f, g=g, operator=operator, theta=5)


class TestSIRT(CCPiTestClass):

    def setUp(self):
        np.random.seed(10)
        # set up matrix, vectordata
        n, m = 50, 50

        A = np.random.uniform(0, 1, (m, n)).astype('float32')
        b = A.dot(np.random.randn(n))

        self.Aop = MatrixOperator(A)
        self.bop = VectorData(b)

        self.ig = self.Aop.domain

        self.initial = self.ig.allocate()

        # set up with linear operator
        self.ig2 = ImageGeometry(3, 4, 5)
        self.initial2 = self.ig2.allocate(0.)
        self.b2 = self.ig2.allocate('random', seed=3)
        self.A2 = IdentityOperator(self.ig2)

    def tearDown(self):
        pass

    def test_set_up(self):

        initial = self.A2.domain_geometry().allocate(0)
        sirt = SIRT(initial=initial, operator=self.A2,
                    data=self.b2, lower=0, upper=1)

        # Test if set_up correctly configures the object
        self.assertTrue(sirt.configured)
        self.assertIsNotNone(sirt.x)
        self.assertIsNotNone(sirt.r)
        self.assertIsNotNone(sirt.constraint)
        self.assertEqual(sirt.constraint.lower, 0)
        self.assertEqual(sirt.constraint.upper, 1)

        constraint = IndicatorBox(lower=0, upper=1)
        sirt = SIRT(initial=None, operator=self.A2,
                    data=self.b2, constraint=constraint)

        # Test if set_up correctly configures the object with constraint
        self.assertTrue(sirt.configured)
        self.assertEqual(sirt.constraint, constraint)

        with self.assertRaises(ValueError) as context:
            sirt = SIRT(initial=None, operator=None, data=self.b2)
            self.assertEqual(str(context.exception),
                             'You must pass an `operator` to the SIRT algorithm')

        with self.assertRaises(ValueError) as context:
            sirt = SIRT(initial=None, operator=self.A2, data=None)
            self.assertEqual(str(context.exception),
                             'You must pass `data` to the SIRT algorithm')
        with self.assertRaises(ValueError) as context:
            sirt = SIRT(initial=None, operator=None, data=None)
            self.assertEqual(str(context.exception),
                             'You must pass an `operator` and `data` to the SIRT algorithm')

        sirt = SIRT(initial=None, operator=self.A2, data=self.b2)
        self.assertTrue(sirt.configured)
        self.assertIsInstance(sirt.x, ImageData)
        self.assertTrue((sirt.x.as_array() == 0).all())

        initial = self.A2.domain_geometry().allocate(1)
        sirt = SIRT(initial=initial, operator=self.A2, data=self.b2)
        self.assertTrue(sirt.configured)
        self.assertIsInstance(sirt.x, ImageData)
        self.assertTrue((sirt.x.as_array() == 1).all())

    def test_update(self):
        # sirt run 5 iterations
        tmp_initial = self.ig.allocate()
        sirt = SIRT(initial=tmp_initial, operator=self.Aop,
                    data=self.bop)
        sirt.run(5)
        x = tmp_initial.copy()
        x_old = tmp_initial.copy()

        for _ in range(5):
            x = x_old + sirt.D * \
                (sirt.operator.adjoint(sirt.M*(sirt.data - sirt.operator.direct(x_old))))
            x_old.fill(x)

        np.testing.assert_allclose(sirt.solution.array, x.array, atol=1e-2)

    def test_update_constraints(self):
        alg = SIRT(initial=self.initial2, operator=self.A2,
                   data=self.b2)
        alg.run(20, verbose=0)
        np.testing.assert_array_almost_equal(alg.x.array, self.b2.array)

        alg = SIRT(initial=self.initial2, operator=self.A2,
                   data=self.b2,  upper=0.3)
        alg.run(20, verbose=0)
        np.testing.assert_almost_equal(alg.solution.max(), 0.3)

        alg = SIRT(initial=self.initial2, operator=self.A2,
                   data=self.b2, lower=0.7)
        alg.run(20, verbose=0)
        np.testing.assert_almost_equal(alg.solution.min(), 0.7)

        alg = SIRT(initial=self.initial2, operator=self.A2, data=self.b2,
                   constraint=IndicatorBox(lower=0.1, upper=0.3))
        alg.run(20, verbose=0)
        np.testing.assert_almost_equal(alg.solution.max(), 0.3)
        np.testing.assert_almost_equal(alg.solution.min(), 0.1)

    def test_SIRT_relaxation_parameter(self):
        tmp_initial = self.ig.allocate()
        alg = SIRT(initial=tmp_initial, operator=self.Aop,
                   data=self.bop)

        with self.assertRaises(ValueError):
            alg.set_relaxation_parameter(0)

        with self.assertRaises(ValueError):
            alg.set_relaxation_parameter(2)

        alg = SIRT(initial=self.initial2, operator=self.A2,
                   data=self.b2)
        alg.set_relaxation_parameter(0.5)

        self.assertEqual(alg.relaxation_parameter, 0.5)

        alg.run(20, verbose=0)
        np.testing.assert_array_almost_equal(alg.x.array, self.b2.array)

        np.testing.assert_almost_equal(0.5 * alg.D.array, alg._Dscaled.array)

    def test_SIRT_nan_inf_values(self):
        Aop_nan_inf = self.Aop
        Aop_nan_inf.A[0:10, :] = 0.
        Aop_nan_inf.A[:, 10:20] = 0.

        tmp_initial = self.ig.allocate()
        sirt = SIRT(initial=tmp_initial, operator=Aop_nan_inf,
                    data=self.bop)

        self.assertFalse(np.any(sirt.M == np.inf))
        self.assertFalse(np.any(sirt.D == np.inf))

    def test_SIRT_remove_nan_or_inf_with_BlockDataContainer(self):
        np.random.seed(10)
        # set up matrix, vectordata
        n, m = 50, 50

        A = np.random.uniform(0, 1, (m, n)).astype('float32')
        b = A.dot(np.random.randn(n))

        A[0:10, :] = 0.
        A[:, 10:20] = 0.
        Aop = BlockOperator(MatrixOperator(A*1), MatrixOperator(A*2))
        bop = BlockDataContainer(VectorData(b*1), VectorData(b*2))

        ig = BlockGeometry(self.ig.copy(), self.ig.copy())
        tmp_initial = ig.allocate()

        sirt = SIRT(initial=tmp_initial, operator=Aop,
                    data=bop)
        for el in sirt.M.containers:
            self.assertFalse(np.any(el == np.inf))

        self.assertFalse(np.any(sirt.D == np.inf))

    def test_SIRT_with_TV(self):
        data = dataexample.SIMPLE_PHANTOM_2D.get(size=(128, 128))
        ig = data.geometry
        A = IdentityOperator(ig)
        constraint = TotalVariation(warm_start=False)
        initial = ig.allocate('random', seed=5)
        sirt = SIRT(initial=initial, operator=A, data=data,
                    constraint=constraint)
        sirt.run(2, verbose=0)
        f = LeastSquares(A, data, c=0.5)
        fista = FISTA(initial=initial, f=f, g=constraint)
        fista.run(100, verbose=0)
        self.assertNumpyArrayAlmostEqual(fista.x.as_array(), sirt.x.as_array())

    def test_SIRT_with_TV_warm_start(self):
        data = dataexample.SIMPLE_PHANTOM_2D.get(size=(128, 128))
        ig = data.geometry
        A = IdentityOperator(ig)
        constraint = 1e6*TotalVariation(warm_start=True, max_iteration=100)
        initial = ig.allocate('random', seed=5)
        sirt = SIRT(initial=initial, operator=A, data=data,
                    constraint=constraint)
        sirt.run(25, verbose=0)

        self.assertNumpyArrayAlmostEqual(
            sirt.x.as_array(), ig.allocate(0.25).as_array(), 3)


class TestSPDHG(CCPiTestClass):
    def setUp(self):
        self.subsets = 10

        data = dataexample.SIMULATED_PARALLEL_BEAM_DATA.get(size=(16, 16))

        partitioned_data = data.partition(self.subsets, 'sequential')
        self.A = BlockOperator(
            *[IdentityOperator(partitioned_data[i].geometry) for i in range(self.subsets)])
        self.A2 = BlockOperator(
            *[IdentityOperator(partitioned_data[i].geometry) for i in range(self.subsets)])

        # block function
        self.F = BlockFunction(*[L2NormSquared(b=partitioned_data[i])
                                for i in range(self.subsets)])
        alpha = 0.025
        self.G = alpha * IndicatorBox(lower=0)



    def test_SPDHG_defaults_and_setters(self):
        # Test SPDHG init with default values
        gamma = 1.
        rho = .99
        spdhg = SPDHG(f=self.F, g=self.G, operator=self.A)

        self.assertListEqual(spdhg._norms, [self.A.get_item(i, 0).norm()
                                            for i in range(self.subsets)])
        self.assertListEqual(spdhg._prob_weights, [
                             1/self.subsets] * self.subsets)
        self.assertEqual(spdhg._sampler._type, 'random_with_replacement')
        self.assertListEqual(
            spdhg.sigma, [rho / ni for ni in spdhg._norms])
        self.assertEqual(spdhg.tau, min([rho*pi / (si * ni**2) for pi, ni,
                                         si in zip(spdhg._prob_weights, spdhg._norms, spdhg.sigma)]))
        self.assertNumpyArrayEqual(
            spdhg.x.as_array(), self.A.domain_geometry().allocate(0).as_array())
        self.assertEqual(spdhg.update_objective_interval, 1)

        # Test SPDHG setters - "from ratio"
        gamma = 3.7
        rho = 5.6
        spdhg.set_step_sizes_from_ratio(gamma, rho)
        self.assertListEqual(
            spdhg.sigma, [gamma * rho / ni for ni in spdhg._norms])
        self.assertEqual(spdhg.tau, min([pi*rho / (si * ni**2) for pi, ni,
                                         si in zip(spdhg._prob_weights, spdhg._norms, spdhg.sigma)]))

        # Test SPDHG setters - set_step_sizes default values for sigma and tau
        gamma = 1.
        rho = .99
        spdhg.set_step_sizes()
        self.assertListEqual(
            spdhg.sigma, [rho / ni for ni in spdhg._norms])
        self.assertEqual(spdhg.tau, min([rho*pi / (si * ni**2) for pi, ni,
                                         si in zip(spdhg._prob_weights, spdhg._norms, spdhg.sigma)]))

        # Test SPDHG setters - set_step_sizes with sigma and tau
        spdhg.set_step_sizes(sigma=[1]*self.subsets, tau=100)
        self.assertListEqual(spdhg.sigma, [1]*self.subsets)
        self.assertEqual(spdhg.tau, 100)

        # Test SPDHG setters - set_step_sizes with sigma
        spdhg.set_step_sizes(sigma=[1]*self.subsets, tau=None)
        self.assertListEqual(spdhg.sigma, [1]*self.subsets)
        self.assertEqual(spdhg.tau, min([(rho*pi / (si * ni**2)) for pi, ni,
                                         si in zip(spdhg._prob_weights, spdhg._norms, spdhg.sigma)]))

        # Test SPDHG setters - set_step_sizes with tau
        spdhg.set_step_sizes(sigma=None, tau=100)
        self.assertListEqual(spdhg.sigma, [
                             gamma * rho*pi / (spdhg.tau*ni**2) for ni, pi in zip(spdhg._norms, spdhg._prob_weights)])
        self.assertEqual(spdhg.tau, 100)

    def test_spdhg_non_default_init(self):
        # Test SPDHG init with non-default values
        spdhg = SPDHG(f=self.F, g=self.G, operator=self.A, sampler=Sampler.random_with_replacement(10, list(np.arange(1, 11)/55.)),
                      initial=self.A.domain_geometry().allocate(1), update_objective_interval=10)

        self.assertListEqual(spdhg._prob_weights,  list(np.arange(1, 11)/55.))
        self.assertNumpyArrayEqual(
            spdhg.x.as_array(), self.A.domain_geometry().allocate(1).as_array())
        self.assertEqual(spdhg.update_objective_interval, 10)

        # Test SPDHG setters - prob_weights and a sampler gives an error
        with self.assertRaises(ValueError):
            spdhg = SPDHG(f=self.F, g=self.G, operator=self.A, prob_weights=[1/(self.subsets-1)]*(
                self.subsets-1)+[0], sampler=Sampler.random_with_replacement(10, list(np.arange(1, 11)/55.)))

        spdhg = SPDHG(f=self.F, g=self.G, operator=self.A, prob_weights=[1/(self.subsets-1)]*(
            self.subsets-1)+[0],  initial=self.A.domain_geometry().allocate(1), update_objective_interval=10)

        self.assertListEqual(spdhg._prob_weights,  [
                             1/(self.subsets-1)]*(self.subsets-1)+[0])
        self.assertEqual(spdhg._sampler._type, 'random_with_replacement')

    def test_spdhg_sampler_gives_too_large_index(self):
        spdhg = SPDHG(f=self.F, g=self.G, operator=self.A, sampler=Sampler.sequential(20),
                      initial=self.A.domain_geometry().allocate(1), update_objective_interval=10)
        with self.assertRaises(IndexError):
            spdhg.run(12)

    def test_spdhg_deprecated_vargs(self):
        spdhg = SPDHG(f=self.F, g=self.G, operator=self.A, norms=[
                      1]*len(self.A), prob=[1/(self.subsets-1)]*(self.subsets-1)+[0])

        self.assertListEqual(self.A.get_norms_as_list(), [1]*len(self.A))
        self.assertListEqual(spdhg._norms, [1]*len(self.A))
        self.assertListEqual(spdhg._sampler.prob_weights, [
                             1/(self.subsets-1)]*(self.subsets-1)+[0])
        self.assertListEqual(spdhg._prob_weights, [
                             1/(self.subsets-1)]*(self.subsets-1)+[0])

        with self.assertRaises(ValueError):
            spdhg = SPDHG(f=self.F, g=self.G, operator=self.A, prob=[1/(self.subsets-1)]*(
                self.subsets-1)+[0], sampler=Sampler.random_with_replacement(10, list(np.arange(1, 11)/55.)))

        with self.assertRaises(ValueError):
            spdhg = SPDHG(f=self.F, g=self.G, operator=self.A,  prob=[1/(self.subsets-1)]*(
                self.subsets-1)+[0], prob_weights=[1/(self.subsets-1)]*(self.subsets-1)+[0])

        with self.assertRaises(ValueError):
            spdhg = SPDHG(f=self.F, g=self.G, operator=self.A, sfdsdf=3,
                          sampler=Sampler.random_with_replacement(10, list(np.arange(1, 11)/55.)))

    def test_spdhg_set_norms(self):

        self.A2.set_norms([1]*len(self.A2))
        spdhg = SPDHG(f=self.F, g=self.G, operator=self.A2)
        self.assertListEqual(spdhg._norms, [1]*len(self.A2))

    def test_spdhg_check_convergence(self):
        spdhg = SPDHG(f=self.F, g=self.G, operator=self.A)

        self.assertTrue(spdhg.check_convergence())

        gamma = 3.7
        rho = 0.9
        spdhg.set_step_sizes_from_ratio(gamma, rho)
        self.assertTrue(spdhg.check_convergence())

        gamma = 3.7
        rho = 100
        spdhg.set_step_sizes_from_ratio(gamma, rho)
        self.assertFalse(spdhg.check_convergence())

        spdhg.set_step_sizes(sigma=[1]*self.subsets, tau=100)
        self.assertFalse(spdhg.check_convergence())

        spdhg.set_step_sizes(sigma=[1]*self.subsets, tau=None)
        self.assertTrue(spdhg.check_convergence())

        spdhg.set_step_sizes(sigma=None, tau=100)
        self.assertTrue(spdhg.check_convergence())

        
    def test_SPDHG_update_method(self):
        sampler = Sampler.sequential(5)
        np.random.seed(5)
        initial = VectorData(np.random.standard_normal(25))
        b = VectorData(np.array(range(25)))
        functions = []
        operators=[]
        for i in range(5):
            diagonal = np.zeros(25)
            diagonal[5*i:5*(i+1)] = 1
            A = MatrixOperator(np.diag(diagonal))
            functions.append(0.5*L2NormSquared(b=A.direct(b)))
            operators.append(A)

        Aop=MatrixOperator(np.diag(np.ones(25)))
         
        g=L1Norm()
        f=BlockFunction(*functions)
        operator=BlockOperator(*operators)
        alg_stochastic = SPDHG(initial=initial, f=BlockFunction(*functions), g=g, operator=operator, sampler=sampler, update_objective_interval=500)
        
        #Test first iteration primal and dual result 
        alg_stochastic.run(1) 
        y_bar_0=operator.range_geometry().allocate(0)
        y_0=operator.range_geometry().allocate(0)
        x_1=g.proximal(initial-alg_stochastic.tau*operator.adjoint(y_bar_0), tau=alg_stochastic.tau) #primal update
        y_1=operator.range_geometry().allocate(0) 
        functions[0].proximal_conjugate(y_0[0]+alg_stochastic.sigma[0]*operators[0].direct(x_1), tau=alg_stochastic.sigma[0], out=y_1[0]) #dual update
        y_bar_1=y_1+alg_stochastic._theta*5*(y_1-y_0) #extrapolation
        
        np.testing.assert_array_almost_equal(alg_stochastic.x.as_array(), x_1.as_array()) #check primal update
        np.testing.assert_array_almost_equal(alg_stochastic._y_old[0].as_array(), y_1[0].as_array()) #check dual update in the first index (should have changed)
        np.testing.assert_array_almost_equal(alg_stochastic._y_old[1].as_array(), y_1[1].as_array())# check dual update in the second index (should not have changed)
        
        #Test second iteration
        alg_stochastic.run(1)

        x_2=g.proximal(x_1-alg_stochastic.tau*operator.adjoint(y_bar_1), tau=alg_stochastic.tau) #primal update
        y_2=y_1.copy()
        functions[1].proximal_conjugate(y_1[1]+alg_stochastic.sigma[1]*operators[1].direct(x_2), tau=alg_stochastic.sigma[1], out=y_2[1]) #dual update
        y_bar_2=y_2+alg_stochastic._theta*5*(y_2-y_1) #extrapolation

        np.testing.assert_array_almost_equal(alg_stochastic.x.as_array(), x_2.as_array()) #check primal update
        np.testing.assert_array_almost_equal(alg_stochastic._y_old[0].as_array(), y_2[0].as_array()) #check dual update in the first index (should not have changed)
        np.testing.assert_array_almost_equal(y_1[0].as_array(), y_2[0].as_array()) #check dual update in the first index (should not have changed)
        np.testing.assert_array_almost_equal(alg_stochastic._y_old[1].as_array(), y_2[1].as_array()) #check dual update in the second index (should have changed)
        
class TestCallbacks(unittest.TestCase):
    class PrintAlgo(Algorithm):
        def __init__(self, update_objective_interval=10, **kwargs):
            super().__init__(update_objective_interval=update_objective_interval, **kwargs)
            self.configured = True

        def update(self):
            self.x = -self.iteration

        def update_objective(self):
            self.loss.append(2 ** getattr(self, 'x', np.nan))

    def test_deprecated_kwargs(self):
        with self.assertWarnsRegex(DeprecationWarning, 'max_iteration'):
            self.PrintAlgo(max_iteration=1000)

        with self.assertWarnsRegex(DeprecationWarning, 'log_file'):
            self.PrintAlgo(log_file="")

    def test_progress(self):
        algo = self.PrintAlgo()

        algo.run(20)
        self.assertListEqual(algo.iterations, [-1, 10, 20])
        algo.run(3, callbacks=[callbacks.TextProgressCallback()])  # upto 23
        self.assertListEqual(algo.iterations, [-1, 10, 20])

        with self.assertWarnsRegex(DeprecationWarning, 'print_interval'):
            algo.run(40, print_interval=2)  # upto 63

        def old_callback(iteration, objective, solution):
            print(f"Called {iteration} {objective} {solution}")

        log = NamedTemporaryFile(delete=False)
        log.close()
        algo.run(20, callbacks=[callbacks.LogfileCallback(
            log.name)], callback=old_callback)
        with open(log.name, 'r') as fd:
            self.assertListEqual(
                ["64/83", "74/83", "83/83", ""],
                [line.lstrip().split(" ", 1)[0] for line in fd.readlines()])
        unlink(log.name)

        its = list(range(10, 90, 10))
        self.assertListEqual([-1] + its, algo.iterations)
        np.testing.assert_array_equal(
            [np.nan] + [2 ** (1-i) for i in its], algo.objective)

    def test_stopiteration(self):
        algo = self.PrintAlgo()
        algo.run(20, callbacks=[])
        self.assertEqual(algo.iteration, 20)

        class EarlyStopping(callbacks.Callback):
            def __call__(self, algorithm: Algorithm):
                if algorithm.iteration >= 15:
                    raise StopIteration

        algo = self.PrintAlgo()
        algo.run(20, callbacks=[EarlyStopping()])
        self.assertEqual(algo.iteration, 15)

    def test_CGLSEarlyStopping(self):
        ig = ImageGeometry(10, 2)
        np.random.seed(2)
        initial = ig.allocate(1.)
        data = ig.allocate('random', seed=3)
        operator = IdentityOperator(ig)
        alg = CGLS(initial=initial, operator=operator, data=data,
                   update_objective_interval=2)

        # Test init
        cb = callbacks.CGLSEarlyStopping(epsilon=0.1, omega=33)
        self.assertEqual(cb.epsilon, 0.1)
        self.assertEqual(cb.omega, 33)

        # Test default values
        cb = callbacks.CGLSEarlyStopping()
        self.assertEqual(cb.epsilon, 1e-6)
        self.assertEqual(cb.omega, 1e6)

        # Tests it doesn't stops iterations if the norm of the current residual is not less than epsilon times the norm of the original residual
        alg.norms = 10
        alg.norms0 = 99
        callbacks.CGLSEarlyStopping(epsilon=0.1)(alg)

        # Test it stops iterations if the norm of the current residual is less than epsilon times the norm of the original residual
        alg.norms = 1
        alg.norms0 = 100
        with self.assertRaises(StopIteration):
            callbacks.CGLSEarlyStopping(epsilon=0.1)(alg)

        # Test it doesn't stop iterations if the norm of x is smaller than omega
        alg.norms = 10
        alg.norms0 = 99
        alg.x = data
        callbacks.CGLSEarlyStopping(epsilon=0.1)(alg)

        # Test it stops iterations if the norm of x is larger than omega
        alg.norms = 10
        alg.norms0 = 99
        alg.x = data
        with self.assertRaises(StopIteration):
            callbacks.CGLSEarlyStopping(epsilon=0.1, omega=0.33)(alg)

    def test_EarlyStoppingObjectiveValue(self):
        ig = ImageGeometry(10, 2)
        np.random.seed(2)
        alg = MagicMock()
        alg.loss = [5]
        callbacks.EarlyStoppingObjectiveValue(0.1)(alg)
        alg.loss = []
        callbacks.EarlyStoppingObjectiveValue(0.1)(alg)
        alg.loss = [5, 10]
        callbacks.EarlyStoppingObjectiveValue(0.1)(alg)
        alg.loss = [5, 5.001]
        with self.assertRaises(StopIteration):
            callbacks.EarlyStoppingObjectiveValue(0.1)(alg)


class TestADMM(unittest.TestCase):
    def setUp(self):
        ig = ImageGeometry(2, 3, 2)
        data = ig.allocate(1, dtype=np.float32)
        noisy_data = data+1

        # TV regularisation parameter
        self.alpha = 1

        self.fidelities = [0.5 * L2NormSquared(b=noisy_data), L1Norm(b=noisy_data),
                           KullbackLeibler(b=noisy_data, backend='numpy')]

        F = self.alpha * MixedL21Norm()
        K = GradientOperator(ig)

        # Compute operator Norm
        normK = K.norm()

        # Primal & dual stepsizes
        self.sigma = 1./normK
        self.tau = 1./normK
        self.F = F
        self.K = K

    def test_ADMM_L2(self):
        self.do_test_with_fidelity(self.fidelities[0])

    def test_ADMM_L1(self):
        self.do_test_with_fidelity(self.fidelities[1])

    def test_ADMM_KL(self):
        self.do_test_with_fidelity(self.fidelities[2])

    def do_test_with_fidelity(self, fidelity):
        alpha = self.alpha
        # F = BlockFunction(alpha * MixedL21Norm(),fidelity)

        G = fidelity
        K = self.K
        F = self.F

        admm = LADMM(f=G, g=F, operator=K, tau=self.tau, sigma=self.sigma,
                     update_objective_interval=10)
        admm.run(1, verbose=0)

        admm_noaxpby = LADMM(f=G, g=F, operator=K, tau=self.tau, sigma=self.sigma,
                             update_objective_interval=10)
        admm_noaxpby.run(1, verbose=0)

        np.testing.assert_array_almost_equal(
            admm.solution.as_array(), admm_noaxpby.solution.as_array())

    def test_compare_with_PDHG(self):
        # Load an image from the CIL gallery.
        data = dataexample.SHAPES.get(size=(64, 64))
        ig = data.geometry
        # Add gaussian noise
        noisy_data = applynoise.gaussian(data, seed=10, var=0.0005)

        # TV regularisation parameter
        alpha = 0.1

        # fidelity = 0.5 * L2NormSquared(b=noisy_data)
        # fidelity = L1Norm(b=noisy_data)
        fidelity = KullbackLeibler(b=noisy_data, backend='numpy')

        # Setup and run the PDHG algorithm
        F = BlockFunction(alpha * MixedL21Norm(), fidelity)
        G = ZeroFunction()
        K = BlockOperator(GradientOperator(ig), IdentityOperator(ig))

        # Compute operator Norm
        normK = K.norm()

        # Primal & dual stepsizes
        sigma = 1./normK
        tau = 1./normK

        pdhg = PDHG(f=F, g=G, operator=K, tau=tau, sigma=sigma,
                    update_objective_interval=10)
        pdhg.run(500, verbose=0)

        sigma = 1
        tau = sigma/normK**2

        admm = LADMM(f=G, g=F, operator=K, tau=tau, sigma=sigma,
                     update_objective_interval=10)
        admm.run(500, verbose=0)
        np.testing.assert_almost_equal(
            admm.solution.array, pdhg.solution.array,  decimal=3)


class Test_PD3O(CCPiTestClass):

    def setUp(self):

        # default test data
        self.data = dataexample.CAMERA.get(size=(32, 32))

    def test_init_and_set_up(self):
        F1 = 0.5 * L2NormSquared(b=self.data)
        H1 = 0.1 * MixedL21Norm()
        operator = GradientOperator(self.data.geometry)
        G1 = IndicatorBox(lower=0)

        algo_pd3o = PD3O(f=F1, g=G1, h=H1, operator=operator)
        self.assertEqual(algo_pd3o.gamma, 0.99*2.0/F1.L) 
        self.assertEqual(algo_pd3o.delta, F1.L/(2.0*operator.norm()**2))
        np.testing.assert_array_equal(
            algo_pd3o.x.array, self.data.geometry.allocate(0).array)
        self.assertTrue(algo_pd3o.configured)

        algo_pd3o = PD3O(f=F1, g=G1, h=H1, operator=operator, gamma=3,
                         delta=43.1, initial=self.data.geometry.allocate('random', seed=3))
        self.assertEqual(algo_pd3o.gamma, 3)
        self.assertEqual(algo_pd3o.delta, 43.1)
        np.testing.assert_array_equal(
            algo_pd3o.x.array, self.data.geometry.allocate('random', seed=3).array)
        self.assertTrue(algo_pd3o.configured)

        with self.assertWarnsRegex(UserWarning, "Please use PDHG instead."):
            algo_pd3o = PD3O(f=ZeroFunction(), g=G1, h=H1,
                             operator=operator, delta=43.1)
            self.assertEqual(algo_pd3o.gamma, 1.0/operator.norm())

    def test_PD3O_PDHG_denoising_1_iteration(self):

        # compare the TV denoising problem using
        # PDHG, PD3O for 1 iteration

        # regularisation parameter
        alpha = 0.1

        # setup PDHG denoising
        F = alpha * MixedL21Norm()
        operator = GradientOperator(self.data.geometry)
        norm_op = operator.norm()
        G = 0.5 * L2NormSquared(b=self.data)
        sigma = 1./norm_op
        tau = 1./norm_op
        pdhg = PDHG(f=F, g=G, operator=operator, tau=tau,
                    sigma=sigma, update_objective_interval=100)
        pdhg.run(1)

        # setup PD3O denoising  (F=ZeroFunction)
        H = alpha * MixedL21Norm()

        F = ZeroFunction()
        gamma = 1./norm_op
        delta = 1./norm_op

        pd3O = PD3O(f=F, g=G, h=H, operator=operator, gamma=gamma, delta=delta,
                    update_objective_interval=100)
        pd3O.run(1)

        # PD3O vs pdhg
        np.testing.assert_allclose(
            pdhg.solution.array, pd3O.solution.array, atol=1e-2)

        # objective values
        np.testing.assert_allclose(
            pdhg.objective[-1], pd3O.objective[-1], atol=1e-2)

 
    def test_PD3O_stochastic_f(self): 
        sampler=Sampler.sequential(3)
        initial = VectorData(np.zeros(21))
        b =  VectorData(np.array(range(1,22)))
        functions=[]
        for i in range(3):
            diagonal=np.zeros(21)
            diagonal[7*i:7*(i+1)]=1
            A=MatrixOperator(np.diag(diagonal))
            functions.append( LeastSquares(A, A.direct(b)))
        
        H1 = 0.1 * L1Norm()
        operator = IdentityOperator(initial.geometry)
        G1 = IndicatorBox(lower=0)


        f=SVRGFunction(functions, sampler, snapshot_update_interval=3)
        algo_pd3o = PD3O(f=f, g=G1, h=H1, operator=operator)
        algo_pd3o.run(1)
        self.assertEqual(f.data_passes[-1], 1)
        self.assertEqual(f.data_passes_indices, [[0,1,2]])
        algo_pd3o.run(1)
        self.assertEqual(f.data_passes[-1], 4/3)
        self.assertEqual(f.data_passes_indices, [[0,1,2], [0]])
        algo_pd3o.run(1)
        self.assertAlmostEqual(f.data_passes[-1], 5/3)
        self.assertEqual(f.data_passes_indices, [[0,1,2], [0], [1]])
        algo_pd3o.run(1)
        self.assertAlmostEqual(f.data_passes[-1], 8/3)
        self.assertEqual(f.data_passes_indices, [[0,1,2], [0], [1], [0,1,2]])
        
        sampler=Sampler.sequential(3)
        f=SVRGFunction(functions, sampler, snapshot_update_interval=5, store_gradients=True)
        algo_pd3o = PD3O(f= 3*(2*f), g=G1, h=H1, operator=operator)
        self.assertTrue(isinstance(3*2*f,ScaledFunction) )
        self.assertEqual(algo_pd3o.f.scalar, 6)
        algo_pd3o.run(1)
        self.assertEqual(f.data_passes[-1], 1)
        self.assertEqual(f.data_passes_indices, [[0,1,2]])
        self.assertNumpyArrayAlmostEqual(f._list_stored_gradients[2].array, f.functions[2].gradient(algo_pd3o.solution).array )
        algo_pd3o.run(1)
        self.assertEqual(f.data_passes[-1], 4/3)
        self.assertEqual(f.data_passes_indices, [[0,1,2], [0]])
        algo_pd3o.run(1)
        self.assertAlmostEqual(f.data_passes[-1], 5/3)
        self.assertEqual(f.data_passes_indices, [[0,1,2], [0], [1]])
        algo_pd3o.run(1)
        self.assertAlmostEqual(f.data_passes[-1], 6/3)
        self.assertEqual(f.data_passes_indices, [[0,1,2], [0], [1], [2]])
        
        
        
        sampler=Sampler.sequential(3)
        f=LSVRGFunction(functions, sampler)
        algo_pd3o = PD3O(f=f, g=G1, h=H1, operator=operator)
        algo_pd3o.run(1)
        self.assertEqual(f.data_passes[-1], 1)
        self.assertEqual(f.data_passes_indices, [[0,1,2]])
            
        sampler=Sampler.sequential(3)
        f=SGFunction(functions, sampler)
        algo_pd3o = PD3O(f=f, g=G1, h=H1, operator=operator)
        algo_pd3o.run(1)
        self.assertEqual(f.data_passes[-1], 1/3)
        self.assertEqual(f.data_passes_indices, [[0]])
        algo_pd3o.run(1)
        self.assertEqual(f.data_passes[-1], 2/3)
        self.assertEqual(f.data_passes_indices, [[0], [1]])
    
        sampler=Sampler.sequential(3)
        f=SAGFunction(functions, sampler)
        algo_pd3o = PD3O(f=f, g=G1, h=H1, operator=operator)
        algo_pd3o.run(1)
        self.assertEqual(f.data_passes[-1], 1/3)
        self.assertEqual(f.data_passes_indices, [[0]])
        algo_pd3o.run(1)
        self.assertEqual(f.data_passes[-1], 2/3)
        self.assertEqual(f.data_passes_indices, [[0], [1]])
        
        sampler=Sampler.sequential(3)
        f=SAGAFunction(functions, sampler)
        algo_pd3o = PD3O(f=f, g=G1, h=H1, operator=operator)
        algo_pd3o.run(1)
        self.assertEqual(f.data_passes[-1], 1/3)
        self.assertEqual(f.data_passes_indices, [[0]])
        algo_pd3o.run(1)
        self.assertEqual(f.data_passes[-1], 2/3)
        self.assertEqual(f.data_passes_indices, [[0], [1]])
        



#  Copyright 2024 United Kingdom Research and Innovation
#  Copyright 2024 The University of Manchester
# 
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
# 
#      http://www.apache.org/licenses/LICENSE-2.0
# 
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# 
# Authors:
# - CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt
# - Daniel Deidda (National Physical Laboratory, UK)
# - Claire Delplancke (Electricite de France, Research and Development)
# - Ashley Gillman (Australian e-Health Res. Ctr., CSIRO, Brisbane, Queensland, Australia)
# - Zeljko Kereta (Department of Computer Science, University College London, UK)
# - Evgueni Ovtchinnikov (STFC - UKRI)
# - Georg Schramm (Department of Imaging and Pathology, Division of Nuclear Medicine, KU Leuven, Leuven, Belgium)

from .ApproximateGradientSumFunction import ApproximateGradientSumFunction
import numpy as np


class SAGFunction(ApproximateGradientSumFunction):

    r"""
    The stochastic average gradient (SAG) function takes a index :math:`i_k` and calculates the approximate gradient of :math:`\sum_{i=0}^{n-1}f_i` at iteration :math:`x_k` as
    
    .. math ::
                \sum_{i=0}^{n-1} g_i^k \qquad \text{where} \qquad g_i^k= \begin{cases}
                                                                            \nabla f_i(x_k), \text{ if } i=i_k\\
                                                                            g_i^{k-1},\text{ otherwise }
                                                                            \end{cases}

        
            
    
    The idea is that by incorporating a memory of previous gradient values the SAG method can achieve a faster convergence rate than black-box stochastic gradient methods. 
    
    Note
    -----
    Compared with the literature, we do not divide by :math:`n`, the number of functions, so that we return an approximate gradient of the whole sum function and not an average gradient.

    Note
    ----------
    Reference: Schmidt, M., Le Roux, N. and Bach, F., 2017. Minimizing finite sums with the stochastic average gradient. Mathematical Programming, 162, pp.83-112. https://doi.org/10.1007/s10107-016-1030-6. 

    Parameters
    -----------
    functions : `list`  of functions
        A list of functions: :math:`f_{0}, f_{1}, ..., f_{n-1}`. Each function is assumed to be smooth with an implemented :func:`~Function.gradient` method. All functions must have the same domain. The number of functions (equivalently the length of the list `n`) must be strictly greater than 1. 
    sampler: An instance of a CIL Sampler class ( :meth:`~optimisation.utilities.sampler`) or of another class which has a `next` function implemented to output integers in :math:`{0,...,n-1}`.
        This sampler is called each time `gradient` is called and sets the internal `function_num` passed to the `approximate_gradient` function.  Default is `Sampler.random_with_replacement(len(functions))`. 
    
    Note
    ------
    
    The user has the option of calling the class method `warm_start_approximate_gradients` after initialising this class. This will compute and store the gradient for each function at an initial point, equivalently setting :math:`g_i^0=\nabla f_i(x_0)` for initial point :math:`x_0`.  If this method is not called, the gradients are initialised with zeros. 

    Note
    ------  

    This function's memory requirements are `n + 3` times the image space, that is with 100 subsets the memory requirement is 103 images, which is huge.
    

    """

    def __init__(self, functions,  sampler=None):
        self._list_stored_gradients = None
        self._full_gradient_at_iterate = None
        self._warm_start_just_done = False
        self._sampled_grad = None
        
        super(SAGFunction, self).__init__(functions, sampler)


        

    def approximate_gradient(self, x, function_num,  out=None):
        """ SAG approximate gradient, calculated at the point :math:`x` and updated using the function index given by `function_num`.  

        Parameters
        ----------
        x : DataContainer (e.g. ImageData object)
            Element in the domain of the `functions`
        function_num: `int` 
            Between 0 and the number of functions in the list  

        """
        
        
        if self._list_stored_gradients is None: # Initialise the stored gradients on the first call of gradient unless using warm start.  
            self._list_stored_gradients = [
                0*x for fi in self.functions]
            self._full_gradient_at_iterate = 0*x
            self._sampled_grad = x.copy()
            self._stochastic_grad_difference = x.copy()
        
        if self.function_num >= self.num_functions or self.function_num<0 : # check the sampler and raise an error if needed
            raise IndexError(f"The sampler has produced the index {self.function_num} which does not match the expected range of available functions to sample from. Please ensure your sampler only selects from [0,1,...,len(functions)-1] ")

            
        # Calculate the gradient of the sampled function at the current iterate 
        self.functions[function_num].gradient(x, out=self._sampled_grad)

        
        # Calculate the difference between the new gradient of the sampled function and the stored one
        self._sampled_grad.sapyb(
            1., self._list_stored_gradients[function_num], -1., out=self._stochastic_grad_difference)

        # Calculate the  approximate gradient
        out = self._update_approx_gradient(out)

        # Update the stored gradients 
        self._list_stored_gradients[function_num].fill(
            self._sampled_grad)
        
        # Calculate the stored full gradient
        self._full_gradient_at_iterate.sapyb(
            1., self._stochastic_grad_difference, 1., out=self._full_gradient_at_iterate)

        return out
    
    def _update_approx_gradient(self, out):
        """Internal function used to differentiate between the SAG and SAGA calculations. This is the SAG approximation: """
        out = self._stochastic_grad_difference.sapyb(  
                1., self._full_gradient_at_iterate, 1., out=out)  

        return out 
    
    def warm_start_approximate_gradients(self, initial):
        r"""A function to warm start SAG or SAGA algorithms by initialising all the gradients at an initial point. Equivalently setting :math:`g_i^0 = \nabla f_i(x_0)` for initial point :math:`x_0`. 
        
        Parameters
        ----------
        initial: DataContainer,
            The initial point to warmstart the calculation
            
        Note
        ------
        When using SAG or SAGA with a deterministic algorithm, you should warm start the SAG-SAGA Function with the same initial point that you initialise the algorithm
        
        
        """
        self._list_stored_gradients = [
            fi.gradient(initial) for fi in self.functions]
        self._full_gradient_at_iterate = np.sum(self._list_stored_gradients)
        self._update_data_passes_indices(list(range(self.num_functions)))
        self._sampled_grad = initial.copy()
        self._stochastic_grad_difference = initial.copy()

    @property
    def data_passes_indices(self): 
        """ The property :code:`data_passes_indices` is a list of lists holding the indices of the functions that are processed in each call of `gradient`. This list is updated each time `gradient` is called by appending a list of the indices of the functions used to calculate the gradient.  
        This is overwritten from the base class to first check to see if the approximate gradient was warm started and, if it was, ensure that the first element of `data_passes_indices` contains each index used to warm start and the index used in the first call to `gradient`. Thus the length of `data_passes_indices` is always equal to the number of calls to `gradient`. 
        """
        ret = self._data_passes_indices[:]  
        if len(ret[0]) == self.num_functions:  
            a = ret.pop(1)  
            ret[0] += a  
        return ret
    
class SAGAFunction(SAGFunction):

    r"""
    SAGA (SAG-Ameliore) is an accelerated version of the stochastic average gradient (SAG) function which takes a index :math:`i_k` and calculates the approximate gradient of :math:`\sum_{i=0}^{n-1}f_i` at iteration :math:`x_k` as
    
    .. math ::
                 n\left(g_{i_k}^{k}-g_{i_k}^{k-1}\right)+\sum_{i=0}^{n-1} g_i^{k-1} \qquad \text{where} \qquad g_i^k= \begin{cases}
                                                                            \nabla f_i(x_k), \text{ if } i=i_k\\
                                                                            g_i^{k-1},\text{ otherwise}
                                                                            \end{cases}
                                                                        
         
    SAGA improves on the theory behind SAG and SVRG, with better theoretical convergence rates. Compared to SAG it is an unbiased estimator. 
    
    Note
    ------
    Compared with the literature, we do not divide by :math:`n`, the number of functions, so that we return an approximate gradient of the whole sum function and not an average gradient.


    Note  
    ------  

    This function's memory requirements are `n + 3` times the image space, that is with 100 subsets the memory requirement is 103 images, which is huge.
    
    
    Note
    ------
    Reference: Defazio, A., Bach, F. and Lacoste-Julien, S., 2014. SAGA: A fast incremental gradient method with support for non-strongly convex composite objectives. Advances in neural information processing systems, 27. https://proceedings.neurips.cc/paper_files/paper/2014/file/ede7e2b6d13a41ddf9f4bdef84fdc737-Paper.pdf
   

    Parameters
    -----------
    functions : `list`  of functions
                A list of functions: :code:`[f_{0}, f_{1}, ..., f_{n-1}]`. Each function is assumed to be smooth function with an implemented :func:`~Function.gradient` method. Each function must have the same domain. The number of functions must be strictly greater than 1. 
    sampler: An instance of one of the :meth:`~optimisation.utilities.sampler` classes which has a `next` function implemented and a `num_indices` property.
        This sampler is called each time gradient is called and  sets the internal `function_num` passed to the `approximate_gradient` function.  The `num_indices` must match the number of functions provided. Default is `Sampler.random_with_replacement(len(functions))`. 
    
    Note
    ----
    The user has the option of calling the class method `warm_start_approximate_gradients` after initialising this class. This will compute and store the gradient for each function at an initial point, equivalently setting :math:`g_i^0=\nabla f_i(x_0)` for initial point :math:`x_0`. If this method is not called, the gradients are initialised with zeros. 

  
    """

    def __init__(self, functions,  sampler=None):
        super(SAGAFunction, self).__init__(functions, sampler)


    def _update_approx_gradient(self, out):
        """Internal function used to differentiate between the SAG and SAGA calculations. This is the SAGA approximation and differs in the constants multiplying the gradients: """

        # Due to the convention that we follow: without the 1/n factor
        out= self._stochastic_grad_difference.sapyb(  
            self.num_functions, self._full_gradient_at_iterate, 1., out)  

        return out 


#  Copyright 2022 United Kingdom Research and Innovation
#  Copyright 2022 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt

import unittest
from utils_projectors import TestCommon_FBP_SIM
from utils import has_astra, has_nvidia, initialise_tests

initialise_tests()

if has_astra:
    from cil.plugins.astra import FBP

def setup_parameters(self):

    self.backend = 'astra'
    self.FBP = FBP
    self.FBP_args={'device':'gpu'}


class Test_Cone3D_FBP_GPU(unittest.TestCase, TestCommon_FBP_SIM):

    @unittest.skipUnless(has_astra and has_nvidia, "Requires ASTRA GPU")
    def setUp(self):
        setup_parameters(self)
        self.Cone3D()
        self.tolerance_fbp = 1e-3
        self.tolerance_fbp_roi = 1e-3


class Test_Cone2D_FBP_GPU(unittest.TestCase, TestCommon_FBP_SIM):

    @unittest.skipUnless(has_astra and has_nvidia, "Requires ASTRA GPU")
    def setUp(self):
        setup_parameters(self)
        self.Cone2D()
        self.tolerance_fbp = 1e-3
        self.tolerance_fbp_roi = 1e-3


class Test_Parallel3D_FBP_GPU(unittest.TestCase, TestCommon_FBP_SIM):

    @unittest.skipUnless(has_astra and has_nvidia, "Requires ASTRA GPU")
    def setUp(self):
        setup_parameters(self)
        self.Parallel3D()
        self.tolerance_fbp = 1e-3
        self.tolerance_fbp_roi = 1e-3


class Test_Parallel2D_FBP_GPU(unittest.TestCase, TestCommon_FBP_SIM):

    @unittest.skipUnless(has_astra and has_nvidia, "Requires ASTRA GPU")
    def setUp(self):
        setup_parameters(self)
        self.Parallel2D()
        self.tolerance_fbp = 1e-3
        self.tolerance_fbp_roi = 1e-3


class Test_Parallel2D_FBP_CPU(unittest.TestCase, TestCommon_FBP_SIM):

    @unittest.skipUnless(has_astra and has_nvidia, "Requires ASTRA")
    def setUp(self):
        setup_parameters(self)
        self.FBP_args['device'] = 'cpu'
        self.Parallel2D()
        self.tolerance_fbp = 1e-3
        self.tolerance_fbp_roi = 1e-3


class Test_Cone3DFlex_FBP_GPU(unittest.TestCase, TestCommon_FBP_SIM):

    @unittest.skipUnless(has_astra and has_nvidia, "Requires ASTRA GPU")
    def setUp(self):
        setup_parameters(self)
        self.Cone3DFlex()
        self.tolerance_fbp = 1e-3
        self.tolerance_fbp_roi = 1e-3


#  Copyright 2020 United Kingdom Research and Innovation
#  Copyright 2020 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt


from cil.framework import AcquisitionGeometry, Processor, ImageData
from cil.framework.labels import AcquisitionType
from cil.plugins.astra.processors.FDK_Flexible import FDK_Flexible
from cil.plugins.astra.utilities import convert_geometry_to_astra_vec_3D, convert_geometry_to_astra
import logging
import astra
import numpy as np

log = logging.getLogger(__name__)


class FBP_Flexible(FDK_Flexible):

    """
    FBP_Flexible Filtered Back Projection performs an FBP reconstruction for 2D and 3D parallel-beam geometries.
    It is able to back-project circular trajectories with 2 PI angular range and equally spaced angular steps.

    This uses the ram-lak filter
    This is a GPU version only

    Parameters
    ----------
    volume_geometry : ImageGeometry
        A description of the area/volume to reconstruct

    sinogram_geometry : AcquisitionGeometry
        A description of the acquisition data

    Example
    -------
    >>> from cil.plugins.astra import FBP_Flexible
    >>> fbp = FBP_Flexible(image_geometry, data.geometry)
    >>> fbp.set_input(data)
    >>> reconstruction = fbp.get_ouput()

    Note
    ----
    ASTRA-toolbox provides limited FBP GPU support for offset/rotated geometries.
    This class creates a pseudo cone-beam geometry in-order to leverage the ASTRA FDK algorithm.
    """

    def __init__(self, volume_geometry,
                       sinogram_geometry):

        super(FBP_Flexible, self).__init__( volume_geometry = volume_geometry, sinogram_geometry = sinogram_geometry)

        #convert parallel geomerty to cone with large source to object
        sino_geom_cone = sinogram_geometry.copy()

        #this catches behaviour modified after CIL 21.3.1
        try:
            sino_geom_cone.config.system.align_reference_frame('cil')
        except:
            sino_geom_cone.config.system.update_reference_frame()

        #reverse ray direction unit-vector direction and extend to inf
        cone_source = -sino_geom_cone.config.system.ray.direction * sino_geom_cone.config.panel.pixel_size[1] * sino_geom_cone.config.panel.num_pixels[1] * 1e6
        detector_position = sino_geom_cone.config.system.detector.position
        detector_direction_x = sino_geom_cone.config.system.detector.direction_x

        if AcquisitionType.DIM2 & sino_geom_cone.dimension:
            tmp = AcquisitionGeometry.create_Cone2D(cone_source, detector_position, detector_direction_x)
        else:
            detector_direction_y = sino_geom_cone.config.system.detector.direction_y
            tmp = AcquisitionGeometry.create_Cone3D(cone_source, detector_position, detector_direction_x, detector_direction_y)

        sino_geom_cone.config.system = tmp.config.system.copy()

        self.vol_geom_astra, self.proj_geom_astra = convert_geometry_to_astra_vec_3D(volume_geometry, sino_geom_cone)

    def check_input(self, dataset):

        if self.sinogram_geometry.channels != 1:
            raise ValueError("Expected input data to be single channel, got {0}"\
                 .format(self.sinogram_geometry.channels))

        if self.sinogram_geometry.geom_type != 'parallel':
            raise ValueError("Expected input data to be parallel beam geometry , got {0}"\
                 .format(self.sinogram_geometry.geom_type))

        return True


class FBP_CPU(Processor):

    """
    FBP_CPU Filtered Back Projection performs an FBP reconstruction for 2D parallel-beam geometries.
    It is able to back-project circular trajectories with 2 PI angular range and equally spaced angular steps.

    This uses the ram-lak filter
    This is a CPU version only

    Parameters
    ----------
    volume_geometry : ImageGeometry
        A description of the area/volume to reconstruct

    sinogram_geometry : AcquisitionGeometry
        A description of the acquisition data

    Example
    -------
    >>> from cil.plugins.astra import FBP_CPU
    >>> fbp = FBP_CPU(image_geometry, data.geometry)
    >>> fbp.set_input(data)
    >>> reconstruction = fbp.get_ouput()
    """


    def __init__(self, volume_geometry,
                       sinogram_geometry):

        super().__init__( volume_geometry = volume_geometry, sinogram_geometry = sinogram_geometry)


    def check_input(self, dataset):

        if self.sinogram_geometry.channels != 1:
            raise ValueError("Expected input data to be single channel, got {0}"\
                 .format(self.sinogram_geometry.channels))

        if self.sinogram_geometry.geom_type != 'parallel':
            raise ValueError("Expected input data to be parallel beam geometry , got {0}"\
                 .format(self.sinogram_geometry.geom_type))

        if not AcquisitionType.DIM2 & self.sinogram_geometry.dimension:
            raise ValueError("Expected input data to be 2D , got {0}"\
                 .format(self.sinogram_geometry.dimension))

        if self.sinogram_geometry.system_description != 'simple':
            log.warning("The ASTRA backend FBP will use simple geometry only. Any configuration offsets or rotations may be ignored.")

        return True
    
    def _set_up(self):
        """
        Configure processor attributes that require the data to setup
        Must set _shape_out
        """
        self._shape_out = self.volume_geometry.shape

    def process(self, out=None):

        # Get DATA
        DATA = self.get_input()

        vol_geom_astra, proj_geom_astra = convert_geometry_to_astra(self.volume_geometry, self.sinogram_geometry)

        rec_id = astra.data2d.create('-vol', vol_geom_astra)
        sinogram_id = astra.data2d.create('-sino', proj_geom_astra, DATA.as_array())
        cfg = astra.astra_dict('FBP')
        cfg['ReconstructionDataId'] = rec_id
        cfg['ProjectionDataId'] = sinogram_id
        cfg['ProjectorId'] = astra.create_projector('line', proj_geom_astra, vol_geom_astra)
        cfg['FilterType'] = 'ram-lak'

        alg_id = astra.algorithm.create(cfg)

        astra.algorithm.run(alg_id)
        arr_out = astra.data2d.get(rec_id)

        astra.data2d.delete(rec_id)
        astra.data2d.delete(sinogram_id)
        astra.algorithm.delete(alg_id)

        arr_out = np.flip(arr_out, 0)
        if out is None:
            out = ImageData(arr_out, deep_copy=False, geometry=self.volume_geometry.copy())
        else:
            out.fill(arr_out)
        return out



from cil.optimisation.algorithms import SPDHG, PDHG, FISTA, APGD, GD, PD3O
from cil.optimisation.functions import L2NormSquared, IndicatorBox, BlockFunction, ZeroFunction, KullbackLeibler, OperatorCompositionFunction, LeastSquares, TotalVariation, MixedL21Norm
from cil.optimisation.operators import BlockOperator, IdentityOperator, MatrixOperator, GradientOperator
from cil.optimisation.utilities import Sampler, BarzilaiBorweinStepSizeRule
from cil.framework import AcquisitionGeometry, BlockDataContainer, BlockGeometry, VectorData, ImageGeometry


from cil.utilities import dataexample
from cil.utilities import noise as applynoise

import numpy as np
import unittest
from testclass import CCPiTestClass

try:
    import cvxpy
    has_cvxpy = True
except ImportError:
    has_cvxpy = False

try:
    import astra
    has_astra = True
    from cil.plugins.astra import ProjectionOperator
except ImportError:
    has_astra = False



class TestAlgorithmConvergence(CCPiTestClass):
    @unittest.skipUnless(has_astra, "cil-astra not available")
    def test_SPDHG_num_subsets_1_astra(self):
        data = dataexample.SIMPLE_PHANTOM_2D.get(size=(10, 10))

        subsets = 1

        ig = data.geometry
        ig.voxel_size_x = 0.1
        ig.voxel_size_y = 0.1

        detectors = ig.shape[0]
        angles = np.linspace(0, np.pi, 90)
        ag = AcquisitionGeometry.create_Parallel2D().set_angles(
            angles, angle_unit='radian').set_panel(detectors, 0.1)
        # Select device
        dev = 'cpu'

        Aop = ProjectionOperator(ig, ag, dev)

        sin = Aop.direct(data)
        partitioned_data = sin.partition(subsets, 'sequential')
        A = BlockOperator(
            *[IdentityOperator(partitioned_data[i].geometry) for i in range(subsets)])

        # block function
        F = BlockFunction(*[L2NormSquared(b=partitioned_data[i])
                            for i in range(subsets)])

        F_phdhg = L2NormSquared(b=partitioned_data[0])
        A_pdhg = IdentityOperator(partitioned_data[0].geometry)

        alpha = 0.025
        G = alpha * IndicatorBox(lower=0)

        spdhg = SPDHG(f=F, g=G, operator=A,  update_objective_interval=10)

        spdhg.run(7)

        pdhg = PDHG(f=F_phdhg, g=G, operator=A_pdhg,
                    update_objective_interval=10)

        pdhg.run(7)

        self.assertNumpyArrayAlmostEqual(
            pdhg.solution.as_array(), spdhg.solution.as_array(), decimal=3)


    @unittest.skipUnless(has_cvxpy, "cvxpy not available")
    def test_SPDHG_toy_example(self):
        sampler = Sampler.random_with_replacement(5, seed=10)
        np.random.seed(10)
        initial = VectorData(np.random.standard_normal(25))
        b = VectorData(np.array(range(25)))
        functions = []
        operators=[]
        for i in range(5):
            diagonal = np.zeros(25)
            diagonal[5*i:5*(i+1)] = 1
            A = MatrixOperator(np.diag(diagonal))
            functions.append(0.5*L2NormSquared(b=A.direct(b)))
            operators.append(A)

        Aop=MatrixOperator(np.diag(np.ones(25)))

        u_cvxpy = cvxpy.Variable(b.shape[0])
        objective = cvxpy.Minimize( 0.5*cvxpy.sum_squares(Aop.A @ u_cvxpy - Aop.direct(b).array))
        p = cvxpy.Problem(objective)
        p.solve(verbose=True, solver=cvxpy.SCS, eps=1e-4)

        g=ZeroFunction()

        alg_stochastic = SPDHG(f=BlockFunction(*functions), g=g, operator=BlockOperator(*operators), sampler=sampler, initial=initial, update_objective_interval=500)
        alg_stochastic.run(200, verbose=0)

        self.assertNumpyArrayAlmostEqual(
            alg_stochastic.x.as_array(), u_cvxpy.value)
        self.assertNumpyArrayAlmostEqual(
            alg_stochastic.x.as_array(), b.as_array(), decimal=6)


    def test_FISTA_Denoising(self):
        # adapted from demo FISTA_Tikhonov_Poisson_Denoising.py in CIL-Demos repository
        data = dataexample.SHAPES.get()
        ig = data.geometry
        ag = ig
        # Create Noisy data with Poisson noise
        scale = 5
        noisy_data = applynoise.poisson(data/scale, seed=10) * scale

        # Regularisation Parameter
        alpha = 10

        # Setup and run the FISTA algorithm
        operator = GradientOperator(ig)
        fid = KullbackLeibler(b=noisy_data)
        reg = OperatorCompositionFunction(alpha * L2NormSquared(), operator)

        initial = ig.allocate()
        fista = FISTA(initial=initial, f=reg, g=fid)
        fista.update_objective_interval = 500
        fista.run(3000, verbose=0)
        rmse = (fista.get_output() - data).norm() / data.as_array().size
        self.assertLess(rmse, 4.2e-4)

    def test_APGD(self):
        ig = ImageGeometry(41, 43, 47)
        initial = ig.allocate(0)
        b = ig.allocate("random")**2
        identity = IdentityOperator(ig)

        f = OperatorCompositionFunction(L2NormSquared(b=b), identity)
        g= IndicatorBox(lower=0)

        apgd = APGD(f=f, g=g, initial=initial, update_objective_interval=100, momentum=0.5)
        apgd.run(500, verbose=0)
        self.assertNumpyArrayAlmostEqual(apgd.solution.as_array(), b.as_array(), decimal=3)


    @unittest.skipUnless(has_cvxpy, "cvxpy not available")
    def test_APGD_dossal_chambolle(self):

        np.random.seed(10)
        n = 100
        m = 50
        A = np.random.normal(0,1, (m, n)).astype('float32')
        b = np.random.normal(0,1, m).astype('float32')
        reg = 0.5

        Aop = MatrixOperator(A)
        bop = VectorData(b)
        ig = Aop.domain

        # cvxpy solutions
        u_cvxpy = cvxpy.Variable(ig.shape[0])
        objective = cvxpy.Minimize( 0.5 * cvxpy.sum_squares(Aop.A @ u_cvxpy - bop.array) + reg/2 * cvxpy.sum_squares(u_cvxpy))
        p = cvxpy.Problem(objective)
        p.solve(verbose=False, solver=cvxpy.SCS, eps=1e-4)

        # default fista
        f = LeastSquares(A=Aop, b=bop, c=0.5)
        g = reg/2*L2NormSquared()
        fista = FISTA(initial=ig.allocate(), f=f, g=g, update_objective_interval=1)
        fista.run(500)
        np.testing.assert_allclose(fista.objective[-1], p.value, atol=1e-3)
        np.testing.assert_allclose(fista.solution.array, u_cvxpy.value, atol=1e-3)

        # fista Dossal Chambolle "On the convergence of the iterates of ”FISTA”
        from cil.optimisation.algorithms.APGD import ScalarMomentumCoefficient
        class DossalChambolle(ScalarMomentumCoefficient):
            def __call__(self, algo=None):
                return (algo.iteration-1)/(algo.iteration+50)
        momentum = DossalChambolle()
        fista_dc = APGD(initial=ig.allocate(), f=f, g=g, update_objective_interval=1, momentum=momentum)
        fista_dc.run(500)
        np.testing.assert_allclose(fista_dc.solution.array, u_cvxpy.value, atol=1e-3)
        np.testing.assert_allclose(fista_dc.solution.array, u_cvxpy.value, atol=1e-3)

    def test_pd3o_convergence(self):
        data = dataexample.CAMERA.get(size=(32, 32))
        # pd30 convergence test using TV denoising

        # regularisation parameter
        alpha = 0.11

        # use TotalVariation from CIL (with Fast Gradient Projection algorithm)
        TV = TotalVariation(max_iteration=40)
        tv_cil = TV.proximal(data, tau=alpha)

        F = alpha * MixedL21Norm()
        operator = GradientOperator(data.geometry)
        norm_op = operator.norm()

        # setup PD3O denoising  (H proximalble and G,F = 1/4 * L2NormSquared)
        H = alpha * MixedL21Norm()
        G = 0.25 * L2NormSquared(b=data)
        F = 0.25 * L2NormSquared(b=data)
        gamma = 2./F.L
        delta = 1./(gamma*norm_op**2)

        pd3O_with_f = PD3O(f=F, g=G, h=H, operator=operator, gamma=gamma, delta=delta,
                           update_objective_interval=100)
        pd3O_with_f.run(800)

        # pd30 vs fista
        np.testing.assert_allclose(
            tv_cil.array, pd3O_with_f.solution.array, atol=1e-2)



    def test_bb_step_size_gd_converge(self):
        np.random.seed(2)
        n = 10
        m = 10
        A = np.array(range(1,n*m+1)).reshape(n,m).astype('float32')
        A = np.diag(1/(np.transpose(A)@np.ones(m)))*A
        x = (np.array(range(n)).astype('float32')-n/2)/n
        b=A@x


        Aop = MatrixOperator(A)
        bop = VectorData(b)
        ig=Aop.domain

        initial = VectorData((np.array(range(n)).astype('float32')-n/2)/(n+1))
        f = LeastSquares(Aop, b=bop, c=2)



        ss_rule=BarzilaiBorweinStepSizeRule(1/f.L, 'short')
        alg = GD(initial=initial, f=f, step_size=ss_rule)
        alg.run(300, verbose=0)
        self.assertNumpyArrayAlmostEqual(alg.x.as_array(), x, decimal=4)

        ss_rule=BarzilaiBorweinStepSizeRule(1/f.L, 'long')
        alg = GD(initial=initial, f=f, step_size=ss_rule)
        alg.run(300, verbose=0)
        self.assertNumpyArrayAlmostEqual(alg.x.as_array(), x, decimal=4)

        ss_rule=BarzilaiBorweinStepSizeRule(1/f.L, 'alternate')
        alg = GD(initial=initial, f=f, step_size=ss_rule)
        alg.run(300, verbose=0)
        self.assertNumpyArrayAlmostEqual(alg.x.as_array(), x, decimal=4)
